"""
API operations on the contents of a data library.
"""
import logging
from typing import Optional

from sqlalchemy.orm.exc import (
    MultipleResultsFound,
    NoResultFound,
)

from galaxy import (
    exceptions,
    managers,
    util,
)
from galaxy.actions.library import (
    LibraryActions,
    validate_path_upload,
)
from galaxy.managers.collections_util import (
    api_payload_to_create_params,
    dictify_dataset_collection_instance,
)
from galaxy.model import (
    ExtendedMetadata,
    ExtendedMetadataIndex,
    LibraryDataset,
    tags,
)
from galaxy.model.base import transaction
from galaxy.structured_app import StructuredApp
from galaxy.web import expose_api
from galaxy.webapps.base.controller import (
    HTTPBadRequest,
    url_for,
    UsesFormDefinitionsMixin,
    UsesLibraryMixinItems,
)
from galaxy.webapps.galaxy.api import BaseGalaxyAPIController

log = logging.getLogger(__name__)


class LibraryContentsController(
    BaseGalaxyAPIController, UsesLibraryMixinItems, UsesFormDefinitionsMixin, LibraryActions
):
    def __init__(self, app: StructuredApp, hda_manager: managers.hdas.HDAManager):
        super().__init__(app)
        self.hda_manager = hda_manager

    @expose_api
    def index(self, trans, library_id, **kwd):
        """
        GET /api/libraries/{library_id}/contents:

        Return a list of library files and folders.

        .. note:: This endpoint is slow for large libraries. Returns all content traversing recursively through all folders.
        .. seealso:: :class:`galaxy.webapps.galaxy.api.FolderContentsController.index` for a faster non-recursive solution

        :param  library_id: the encoded id of the library
        :type   library_id: str

        :returns:   list of dictionaries of the form:

            * id:   the encoded id of the library item
            * name: the 'library path'
                or relationship of the library item to the root
            * type: 'file' or 'folder'
            * url:  the url to get detailed information on the library item

        :rtype:     list

        :raises:  MalformedId, InconsistentDatabase, RequestParameterInvalidException, InternalServerError
        """
        rval = []
        current_user_roles = trans.get_current_user_roles()

        def traverse(folder):
            admin = trans.user_is_admin
            rval = []
            for subfolder in folder.active_folders:
                if not admin:
                    can_access, folder_ids = trans.app.security_agent.check_folder_contents(
                        trans.user, current_user_roles, subfolder
                    )
                if (admin or can_access) and not subfolder.deleted:
                    subfolder.api_path = f"{folder.api_path}/{subfolder.name}"
                    subfolder.api_type = "folder"
                    rval.append(subfolder)
                    rval.extend(traverse(subfolder))
            for ld in folder.datasets:
                if not admin:
                    can_access = trans.app.security_agent.can_access_dataset(
                        current_user_roles, ld.library_dataset_dataset_association.dataset
                    )
                if (admin or can_access) and not ld.deleted:
                    ld.api_path = f"{folder.api_path}/{ld.name}"
                    ld.api_type = "file"
                    rval.append(ld)
            return rval

        decoded_library_id = self.decode_id(library_id)
        try:
            library = (
                trans.sa_session.query(trans.app.model.Library)
                .filter(trans.app.model.Library.table.c.id == decoded_library_id)
                .one()
            )
        except MultipleResultsFound:
            raise exceptions.InconsistentDatabase("Multiple libraries found with the same id.")
        except NoResultFound:
            raise exceptions.RequestParameterInvalidException("No library found with the id provided.")
        except Exception as e:
            raise exceptions.InternalServerError(f"Error loading from the database.{util.unicodify(e)}")
        if not (trans.user_is_admin or trans.app.security_agent.can_access_library(current_user_roles, library)):
            raise exceptions.RequestParameterInvalidException("No library found with the id provided.")
        encoded_id = f"F{trans.security.encode_id(library.root_folder.id)}"
        # appending root folder
        rval.append(
            dict(
                id=encoded_id,
                type="folder",
                name="/",
                url=url_for("library_content", library_id=library_id, id=encoded_id),
            )
        )
        library.root_folder.api_path = ""
        # appending all other items in the library recursively
        for content in traverse(library.root_folder):
            encoded_id = trans.security.encode_id(content.id)
            if content.api_type == "folder":
                encoded_id = f"F{encoded_id}"
            rval.append(
                dict(
                    id=encoded_id,
                    type=content.api_type,
                    name=content.api_path,
                    url=url_for(
                        "library_content",
                        library_id=library_id,
                        id=encoded_id,
                    ),
                )
            )
        return rval

    @expose_api
    def show(self, trans, id, library_id, **kwd):
        """
        GET /api/libraries/{library_id}/contents/{id}

        Returns information about library file or folder.

        :param  id:         the encoded id of the library item to return
        :type   id:         str

        :param  library_id: the encoded id of the library that contains this item
        :type   library_id: str

        :returns:   detailed library item information
        :rtype:     dict

        .. seealso::
            :func:`galaxy.model.LibraryDataset.to_dict` and
            :attr:`galaxy.model.LibraryFolder.dict_element_visible_keys`
        """
        class_name, content_id = self._decode_library_content_id(id)
        if class_name == "LibraryFolder":
            content = self.get_library_folder(trans, content_id, check_ownership=False, check_accessible=True)
            rval = content.to_dict(view="element", value_mapper={"id": trans.security.encode_id})
            rval["id"] = f"F{str(rval['id'])}"
            if rval["parent_id"] is not None:  # This can happen for root folders.
                rval["parent_id"] = f"F{str(trans.security.encode_id(rval['parent_id']))}"
            rval["parent_library_id"] = trans.security.encode_id(rval["parent_library_id"])
        else:
            content = self.get_library_dataset(trans, content_id, check_ownership=False, check_accessible=True)
            rval = content.to_dict(view="element")
            rval["id"] = trans.security.encode_id(rval["id"])
            rval["ldda_id"] = trans.security.encode_id(rval["ldda_id"])
            rval["folder_id"] = f"F{str(trans.security.encode_id(rval['folder_id']))}"
            rval["parent_library_id"] = trans.security.encode_id(rval["parent_library_id"])

            tag_manager = tags.GalaxyTagHandler(trans.sa_session)
            rval["tags"] = tag_manager.get_tags_list(content.library_dataset_dataset_association.tags)
        return rval

    @expose_api
    def create(self, trans, library_id, payload, **kwd):
        """
        POST /api/libraries/{library_id}/contents:

        Create a new library file or folder.

        To copy an HDA into a library send ``create_type`` of 'file' and
        the HDA's encoded id in ``from_hda_id`` (and optionally ``ldda_message``).

        To copy an HDCA into a library send ``create_type`` of 'file' and
        the HDCA's encoded id in ``from_hdca_id`` (and optionally ``ldda_message``).

        :type   library_id: str
        :param  library_id: the encoded id of the library where to create the new item
        :type   payload:    dict
        :param  payload:    dictionary structure containing:

            * folder_id:    the encoded id of the parent folder of the new item
            * create_type:  the type of item to create ('file', 'folder' or 'collection')
            * from_hda_id:  (optional, only if create_type is 'file') the
                encoded id of an accessible HDA to copy into the library
            * ldda_message: (optional) the new message attribute of the LDDA created
            * extended_metadata: (optional) sub-dictionary containing any extended
                metadata to associate with the item
            * upload_option: (optional) one of 'upload_file' (default), 'upload_directory' or 'upload_paths'
            * server_dir: (optional, only if upload_option is
                'upload_directory') relative path of the subdirectory of Galaxy
                ``library_import_dir`` (if admin) or ``user_library_import_dir``
                (if non-admin) to upload. All and only the files (i.e.
                no subdirectories) contained in the specified directory will be
                uploaded.
            * filesystem_paths: (optional, only if upload_option is
                'upload_paths' and the user is an admin) file paths on the
                Galaxy server to upload to the library, one file per line
            * link_data_only: (optional, only when upload_option is
                'upload_directory' or 'upload_paths') either 'copy_files'
                (default) or 'link_to_files'. Setting to 'link_to_files'
                symlinks instead of copying the files
            * name: (optional, only if create_type is 'folder') name of the
                folder to create
            * description: (optional, only if create_type is 'folder')
                description of the folder to create
            * tag_using_filenames: (optional)
                create tags on datasets using the file's original name
            * tags: (optional)
                create the given list of tags on datasets

        :returns:   a dictionary describing the new item unless ``from_hdca_id`` is supplied,
                    in that case a list of such dictionaries is returned.
        :rtype:     object
        """
        if trans.user_is_bootstrap_admin:
            raise exceptions.RealUserRequiredException("Only real users can create a new library file or folder.")
        if "create_type" not in payload:
            raise exceptions.RequestParameterMissingException("Missing required 'create_type' parameter.")
        create_type = payload.pop("create_type")
        if create_type not in ("file", "folder", "collection"):
            raise exceptions.RequestParameterInvalidException(
                f"Invalid value for 'create_type' parameter ( {create_type} ) specified."
            )
        if "upload_option" in payload and payload["upload_option"] not in (
            "upload_file",
            "upload_directory",
            "upload_paths",
        ):
            raise exceptions.RequestParameterInvalidException(
                f"Invalid value for 'upload_option' parameter ( {payload['upload_option']} ) specified."
            )
        if "folder_id" not in payload:
            raise exceptions.RequestParameterMissingException("Missing required 'folder_id' parameter.")
        folder_id = payload.pop("folder_id")
        _, folder_id = self._decode_library_content_id(folder_id)
        folder_id = trans.security.decode_id(folder_id)
        # security is checked in the downstream controller
        parent = self.get_library_folder(trans, folder_id, check_ownership=False, check_accessible=False)
        # The rest of the security happens in the library_common controller.

        payload["tag_using_filenames"] = util.string_as_bool(payload.get("tag_using_filenames", None))
        payload["tags"] = util.listify(payload.get("tags", None))

        # are we copying an HDA to the library folder?
        #   we'll need the id and any message to attach, then branch to that private function
        from_hda_id, from_hdca_id, ldda_message = (
            payload.pop("from_hda_id", None),
            payload.pop("from_hdca_id", None),
            payload.pop("ldda_message", ""),
        )
        if create_type == "file":
            if from_hda_id:
                return self._copy_hda_to_library_folder(
                    trans, self.hda_manager, self.decode_id(from_hda_id), folder_id, ldda_message
                )
            if from_hdca_id:
                return self._copy_hdca_to_library_folder(
                    trans, self.hda_manager, self.decode_id(from_hdca_id), folder_id, ldda_message
                )

        # check for extended metadata, store it and pop it out of the param
        # otherwise sanitize_param will have a fit
        ex_meta_payload = payload.pop("extended_metadata", None)

        # Now create the desired content object, either file or folder.
        if create_type == "file":
            status, output = self._upload_library_dataset(trans, folder_id, **payload)
        elif create_type == "folder":
            status, output = self._create_folder(trans, folder_id, **payload)
        elif create_type == "collection":
            # Not delegating to library_common, so need to check access to parent
            # folder here.
            self.check_user_can_add_to_library_item(trans, parent, check_accessible=True)
            create_params = api_payload_to_create_params(payload)
            create_params["parent"] = parent
            dataset_collection_manager = trans.app.dataset_collection_manager
            dataset_collection_instance = dataset_collection_manager.create(**create_params)
            return [
                dictify_dataset_collection_instance(
                    dataset_collection_instance, security=trans.security, url_builder=trans.url_builder, parent=parent
                )
            ]
        if status != 200:
            trans.response.status = status
            return output
        else:
            rval = []
            for v in output.values():
                if ex_meta_payload is not None:
                    # If there is extended metadata, store it, attach it to the dataset, and index it
                    ex_meta = ExtendedMetadata(ex_meta_payload)
                    trans.sa_session.add(ex_meta)
                    v.extended_metadata = ex_meta
                    trans.sa_session.add(v)
                    with transaction(trans.sa_session):
                        trans.sa_session.commit()
                    for path, value in self._scan_json_block(ex_meta_payload):
                        meta_i = ExtendedMetadataIndex(ex_meta, path, value)
                        trans.sa_session.add(meta_i)
                    with transaction(trans.sa_session):
                        trans.sa_session.commit()
                if type(v) == trans.app.model.LibraryDatasetDatasetAssociation:
                    v = v.library_dataset
                encoded_id = trans.security.encode_id(v.id)
                if create_type == "folder":
                    encoded_id = f"F{encoded_id}"
                rval.append(
                    dict(
                        id=encoded_id, name=v.name, url=url_for("library_content", library_id=library_id, id=encoded_id)
                    )
                )
            return rval

    def _upload_library_dataset(self, trans, folder_id: int, **kwd):
        replace_dataset: Optional[LibraryDataset] = None
        upload_option = kwd.get("upload_option", "upload_file")
        dbkey = kwd.get("dbkey", "?")
        if isinstance(dbkey, list):
            last_used_build = dbkey[0]
        else:
            last_used_build = dbkey
        roles = kwd.get("roles", "")
        is_admin = trans.user_is_admin
        current_user_roles = trans.get_current_user_roles()
        folder = trans.sa_session.query(trans.app.model.LibraryFolder).get(folder_id)
        self._check_access(trans, is_admin, folder, current_user_roles)
        self._check_add(trans, is_admin, folder, current_user_roles)
        library = folder.parent_library
        if folder and last_used_build in ["None", None, "?"]:
            last_used_build = folder.genome_build
        error = False
        if upload_option == "upload_paths":
            validate_path_upload(trans)  # Duplicate check made in _upload_dataset.
        elif roles:
            # Check to see if the user selected roles to associate with the DATASET_ACCESS permission
            # on the dataset that would cause accessibility issues.
            vars = dict(DATASET_ACCESS_in=roles)
            permissions, in_roles, error, message = trans.app.security_agent.derive_roles_from_access(
                trans, library.id, "api", library=True, **vars
            )
        if error:
            return 400, message
        else:
            created_outputs_dict = self._upload_dataset(
                trans, folder_id=folder.id, replace_dataset=replace_dataset, **kwd
            )
            if created_outputs_dict:
                if type(created_outputs_dict) == str:
                    return 400, created_outputs_dict
                elif type(created_outputs_dict) == tuple:
                    return created_outputs_dict[0], created_outputs_dict[1]
                return 200, created_outputs_dict
            else:
                return 400, "Upload failed"

    def _scan_json_block(self, meta, prefix=""):
        """
        Scan a json style data structure, and emit all fields and their values.
        Example paths

        Data
        { "data" : [ 1, 2, 3 ] }

        Path:
        /data == [1,2,3]

        /data/[0] == 1

        """
        if isinstance(meta, dict):
            for a in meta:
                yield from self._scan_json_block(meta[a], f"{prefix}/{a}")
        elif isinstance(meta, list):
            for i, a in enumerate(meta):
                yield from self._scan_json_block(a, prefix + "[%d]" % (i))
        else:
            # BUG: Everything is cast to string, which can lead to false positives
            # for cross type comparisions, ie "True" == True
            yield prefix, (f"{meta}").encode()

    @expose_api
    def update(self, trans, id, library_id, payload, **kwd):
        """
        PUT /api/libraries/{library_id}/contents/{id}

        Create an ImplicitlyConvertedDatasetAssociation.

        .. seealso:: :class:`galaxy.model.ImplicitlyConvertedDatasetAssociation`

        :type   id:         str
        :param  id:         the encoded id of the library item to return
        :type   library_id: str
        :param  library_id: the encoded id of the library that contains this item
        :type   payload:    dict
        :param  payload:    dictionary structure containing::
            'converted_dataset_id':

        :rtype:     None
        :returns:   None
        """
        if "converted_dataset_id" in payload:
            converted_id = payload.pop("converted_dataset_id")
            content = self.get_library_dataset(trans, id, check_ownership=False, check_accessible=False)
            content_conv = self.get_library_dataset(trans, converted_id, check_ownership=False, check_accessible=False)
            assoc = trans.app.model.ImplicitlyConvertedDatasetAssociation(
                parent=content.library_dataset_dataset_association,
                dataset=content_conv.library_dataset_dataset_association,
                file_type=content_conv.library_dataset_dataset_association.extension,
                metadata_safe=True,
            )
            trans.sa_session.add(assoc)
            with transaction(trans.sa_session):
                trans.sa_session.commit()

    def _decode_library_content_id(self, content_id):
        if len(content_id) % 16 == 0:
            return "LibraryDataset", content_id
        elif content_id.startswith("F"):
            return "LibraryFolder", content_id[1:]
        else:
            raise HTTPBadRequest(f"Malformed library content id ( {str(content_id)} ) specified, unable to decode.")

    @expose_api
    def delete(self, trans, library_id, id, **kwd):
        """
        DELETE /api/libraries/{library_id}/contents/{id}

        Delete the LibraryDataset with the given ``id``.

        :type   id:     str
        :param  id:     the encoded id of the library dataset to delete
        :type   kwd:    dict
        :param  kwd:    (optional) dictionary structure containing:

            * payload:     a dictionary itself containing:
                * purge:   if True, purge the LD

        :rtype:     dict
        :returns:   an error object if an error occurred or a dictionary containing:
            * id:         the encoded id of the library dataset,
            * deleted:    if the library dataset was marked as deleted,
            * purged:     if the library dataset was purged
        """
        purge = False
        if kwd.get("payload", None):
            purge = util.string_as_bool(kwd["payload"].get("purge", False))

        rval = {"id": id}
        try:
            ld = self.get_library_dataset(trans, id, check_ownership=False, check_accessible=True)
            user_is_admin = trans.user_is_admin
            can_modify = trans.app.security_agent.can_modify_library_item(trans.user.all_roles(), ld)
            log.debug("is_admin: %s, can_modify: %s", user_is_admin, can_modify)
            if not (user_is_admin or can_modify):
                trans.response.status = 403
                rval.update({"error": "Unauthorized to delete or purge this library dataset"})
                return rval

            ld.deleted = True
            if purge:
                ld.purged = True
                trans.sa_session.add(ld)
                with transaction(trans.sa_session):
                    trans.sa_session.commit()

                # TODO: had to change this up a bit from Dataset.user_can_purge
                dataset = ld.library_dataset_dataset_association.dataset
                no_history_assoc = len(dataset.history_associations) == len(dataset.purged_history_associations)
                no_library_assoc = dataset.library_associations == [ld.library_dataset_dataset_association]
                can_purge_dataset = not dataset.purged and no_history_assoc and no_library_assoc

                if can_purge_dataset:
                    try:
                        ld.library_dataset_dataset_association.dataset.full_delete()
                        trans.sa_session.add(ld.dataset)
                    except Exception:
                        pass
                    # flush now to preserve deleted state in case of later interruption
                    with transaction(trans.sa_session):
                        trans.sa_session.commit()
                rval["purged"] = True
            with transaction(trans.sa_session):
                trans.sa_session.commit()
            rval["deleted"] = True

        except exceptions.httpexceptions.HTTPInternalServerError:
            log.exception("Library_contents API, delete: uncaught HTTPInternalServerError: %s, %s", id, str(kwd))
            raise
        except exceptions.httpexceptions.HTTPException:
            raise
        except Exception as exc:
            log.exception("library_contents API, delete: uncaught exception: %s, %s", id, str(kwd))
            trans.response.status = 500
            rval.update({"error": util.unicodify(exc)})
        return rval



from typing import (
    List,
    Optional,
)

from pydantic import BaseModel
from sqlalchemy import (
    and_,
    cast,
    Integer,
)

from galaxy.model.scoped_session import install_model_scoped_session
from galaxy.model.tool_shed_install import ToolShedRepository
from galaxy.schema.fields import DecodedDatabaseIdField
from galaxy.schema.schema import (
    CheckForUpdatesResponse,
    InstalledToolShedRepository,
)
from galaxy.tool_shed.util.repository_util import (
    check_for_updates,
    get_tool_shed_repository_by_decoded_id,
)
from galaxy.util.tool_shed.tool_shed_registry import Registry
from galaxy.web import url_for


class InstalledToolShedRepositoryIndexRequest(BaseModel):
    name: Optional[str] = None
    owner: Optional[str] = None
    changeset: Optional[str] = None
    deleted: Optional[bool] = None
    uninstalled: Optional[bool] = None


class ToolShedRepositoriesService:
    def __init__(
        self,
        install_model_context: install_model_scoped_session,
        tool_shed_registry: Registry,
    ):
        self._install_model_context = install_model_context
        self._tool_shed_registry = tool_shed_registry

    def index(self, request: InstalledToolShedRepositoryIndexRequest) -> List[InstalledToolShedRepository]:
        clause_list = []
        if request.name is not None:
            clause_list.append(ToolShedRepository.table.c.name == request.name)
        if request.owner is not None:
            clause_list.append(ToolShedRepository.table.c.owner == request.owner)
        if request.changeset is not None:
            clause_list.append(ToolShedRepository.table.c.changeset_revision == request.changeset)
        if request.deleted is not None:
            clause_list.append(ToolShedRepository.table.c.deleted == request.deleted)
        if request.uninstalled is not None:
            clause_list.append(ToolShedRepository.table.c.uninstalled == request.uninstalled)
        query = (
            self._install_model_context.query(ToolShedRepository)
            .order_by(ToolShedRepository.table.c.name)
            .order_by(cast(ToolShedRepository.ctx_rev, Integer).desc())
        )
        if len(clause_list) > 0:
            query = query.filter(and_(*clause_list))
        index = []
        for repository in query.all():
            index.append(self._show(repository))
        return index

    def show(self, repository_id: DecodedDatabaseIdField) -> InstalledToolShedRepository:
        tool_shed_repository = get_tool_shed_repository_by_decoded_id(self._install_model_context, int(repository_id))
        return self._show(tool_shed_repository)

    def check_for_updates(self, repository_id: Optional[int]) -> CheckForUpdatesResponse:
        message, status = check_for_updates(self._tool_shed_registry, self._install_model_context, repository_id)
        return CheckForUpdatesResponse(message=message, status=status)

    def _show(self, tool_shed_repository: ToolShedRepository) -> InstalledToolShedRepository:
        tool_shed_repository_dict = tool_shed_repository.as_dict()
        encoded_id = DecodedDatabaseIdField.encode(tool_shed_repository.id)
        tool_shed_repository_dict["id"] = encoded_id
        tool_shed_repository_dict["error_message"] = tool_shed_repository.error_message or ""
        tool_shed_repository_dict["url"] = url_for("tool_shed_repositories", id=encoded_id)
        return InstalledToolShedRepository(**tool_shed_repository_dict)



"""
API operations on Role objects.
"""
import logging

from fastapi import Body

from galaxy.managers.context import ProvidesUserContext
from galaxy.managers.roles import RoleManager
from galaxy.schema.fields import DecodedDatabaseIdField
from galaxy.schema.schema import (
    RoleDefinitionModel,
    RoleListResponse,
    RoleModelResponse,
)
from galaxy.webapps.base.controller import url_for
from galaxy.webapps.galaxy.api import (
    depends,
    DependsOnTrans,
    Router,
)

log = logging.getLogger(__name__)


# Empty paths (e.g. /api/roles) only work if a prefix is defined right here.
# https://github.com/tiangolo/fastapi/pull/415/files
router = Router(tags=["roles"])


def role_to_model(role):
    item = role.to_dict(view="element")
    role_id = DecodedDatabaseIdField.encode(role.id)
    item["url"] = url_for("role", id=role_id)
    return RoleModelResponse(**item)


@router.cbv
class FastAPIRoles:
    role_manager: RoleManager = depends(RoleManager)

    @router.get("/api/roles")
    def index(self, trans: ProvidesUserContext = DependsOnTrans) -> RoleListResponse:
        roles = self.role_manager.list_displayable_roles(trans)
        return RoleListResponse(__root__=[role_to_model(r) for r in roles])

    @router.get("/api/roles/{id}")
    def show(self, id: DecodedDatabaseIdField, trans: ProvidesUserContext = DependsOnTrans) -> RoleModelResponse:
        role = self.role_manager.get(trans, id)
        return role_to_model(role)

    @router.post("/api/roles", require_admin=True)
    def create(
        self, trans: ProvidesUserContext = DependsOnTrans, role_definition_model: RoleDefinitionModel = Body(...)
    ) -> RoleModelResponse:
        role = self.role_manager.create_role(trans, role_definition_model)
        return role_to_model(role)



"""
API for searching Galaxy Datasets
"""
import logging

from galaxy import (
    model,
    web,
)
from galaxy.exceptions import ItemAccessibilityException
from galaxy.managers.context import ProvidesUserContext
from galaxy.model.search import GalaxySearchEngine
from galaxy.util import unicodify
from galaxy.webapps.base.controller import SharableItemSecurityMixin
from . import BaseGalaxyAPIController

log = logging.getLogger(__name__)


class SearchController(BaseGalaxyAPIController, SharableItemSecurityMixin):
    @web.legacy_expose_api
    def create(self, trans: ProvidesUserContext, payload: dict, **kwd):
        """
        POST /api/search
        Do a search of the various elements of Galaxy.
        """
        query_txt = payload.get("query", None)
        out = []
        if query_txt is not None:
            se = GalaxySearchEngine()
            try:
                query = se.query(query_txt)
            except Exception as e:
                return {"error": unicodify(e)}
            if query is not None:
                query.decode_query_ids(trans)
                current_user_roles = trans.get_current_user_roles()
                try:
                    results = query.process(trans)
                except Exception as e:
                    return {"error": unicodify(e)}
                for item in results:
                    append = False
                    if trans.user_is_admin:
                        append = True
                    if not append:
                        if type(item) in [
                            model.LibraryFolder,
                            model.LibraryDatasetDatasetAssociation,
                            model.LibraryDataset,
                        ]:
                            if trans.app.security_agent.can_access_library_item(
                                trans.get_current_user_roles(), item, trans.user
                            ):
                                append = True
                        elif type(item) in [model.Job]:
                            if item.used_id == trans.user or trans.user_is_admin:
                                append = True
                        elif type(item) in [model.Page, model.StoredWorkflow]:
                            try:
                                if self.security_check(trans, item, False, True):
                                    append = True
                            except ItemAccessibilityException:
                                append = False
                        elif type(item) in [model.PageRevision]:
                            try:
                                if self.security_check(trans, item.page, False, True):
                                    append = True
                            except ItemAccessibilityException:
                                append = False
                        elif hasattr(item, "dataset"):
                            if trans.app.security_agent.can_access_dataset(current_user_roles, item.dataset):
                                append = True

                    if append:
                        row = query.item_to_api_value(item)
                        out.append(self.encode_all_ids(trans, row, True))
        return {"results": out}



"""
Contains the user interface in the Universe class
"""

from sqlalchemy import false

from galaxy import (
    util,
    web,
)
from galaxy.model.base import transaction
from galaxy.webapps.base.controller import (
    BaseUIController,
    UsesFormDefinitionsMixin,
)


class User(BaseUIController, UsesFormDefinitionsMixin):
    @web.expose
    @web.require_login()
    @web.require_admin
    def index(self, trans, cntrller, **kwd):
        return self.get_all_users(trans)

    @web.expose
    @web.require_login()
    @web.require_admin
    def admin_api_keys(self, trans, uid, **kwd):
        params = util.Params(kwd)
        uid = params.get("uid", uid)
        new_key = trans.app.model.APIKeys()
        new_key.user_id = trans.security.decode_id(uid)
        new_key.key = trans.app.security.get_new_guid()
        trans.sa_session.add(new_key)
        with transaction(trans.sa_session):
            trans.sa_session.commit()
        return self.get_all_users(trans)

    @web.expose
    @web.require_login()
    @web.require_admin
    def all_users(self, trans, **kwd):
        return self.get_all_users(trans)

    @web.json
    def get_all_users(self, trans):
        users = []
        for user in (
            trans.sa_session.query(trans.app.model.User)
            .filter(trans.app.model.User.table.c.deleted == false())
            .order_by(trans.app.model.User.table.c.email)
        ):
            uid = int(user.id)
            userkey = ""
            for api_user in trans.sa_session.query(trans.app.model.APIKeys).filter(
                trans.app.model.APIKeys.user_id == uid
            ):
                userkey = api_user.key
            users.append({"uid": trans.security.encode_id(uid), "email": user.email, "key": userkey})
        return users



import logging
import os
from urllib.parse import (
    quote_plus,
    unquote_plus,
)

import paste.httpexceptions
from markupsafe import escape

from galaxy import (
    datatypes,
    util,
    web,
)
from galaxy.datatypes.data import DatatypeConverterNotFoundException
from galaxy.datatypes.display_applications.util import (
    decode_dataset_user,
    encode_dataset_user,
)
from galaxy.datatypes.sniff import guess_ext
from galaxy.exceptions import RequestParameterInvalidException
from galaxy.managers.hdas import (
    HDADeserializer,
    HDAManager,
)
from galaxy.managers.histories import HistoryManager
from galaxy.model import Dataset
from galaxy.model.base import transaction
from galaxy.model.item_attrs import (
    UsesAnnotations,
    UsesItemRatings,
)
from galaxy.structured_app import StructuredApp
from galaxy.util import (
    inflector,
    sanitize_text,
)
from galaxy.util.sanitize_html import sanitize_html
from galaxy.util.zipstream import ZipstreamWrapper
from galaxy.web import form_builder
from galaxy.web.framework.helpers import iff
from galaxy.webapps.base.controller import (
    BaseUIController,
    ERROR,
    SUCCESS,
    url_for,
    UsesExtendedMetadataMixin,
)
from galaxy.webapps.galaxy.services.datasets import DatasetsService
from ..api import depends

log = logging.getLogger(__name__)

comptypes = []

try:
    import zlib  # noqa: F401

    comptypes.append("zip")
except ImportError:
    pass


class DatasetInterface(BaseUIController, UsesAnnotations, UsesItemRatings, UsesExtendedMetadataMixin):
    history_manager: HistoryManager = depends(HistoryManager)
    hda_manager: HDAManager = depends(HDAManager)
    hda_deserializer: HDADeserializer = depends(HDADeserializer)
    service: DatasetsService = depends(DatasetsService)

    def __init__(self, app: StructuredApp):
        super().__init__(app)

    def _can_access_dataset(self, trans, dataset_association, allow_admin=True, additional_roles=None):
        roles = trans.get_current_user_roles()
        if additional_roles:
            roles = roles + additional_roles
        return (allow_admin and trans.user_is_admin) or trans.app.security_agent.can_access_dataset(
            roles, dataset_association.dataset
        )

    @web.expose
    def default(self, trans, dataset_id=None, **kwd):
        return "This link may not be followed from within Galaxy."

    @web.expose_api_raw_anonymous_and_sessionless
    def get_metadata_file(self, trans, hda_id, metadata_name, **kwd):
        """Allows the downloading of metadata files associated with datasets (eg. bai index for bam files)"""
        # Backward compatibility with legacy links, should use `/api/datasets/{hda_id}/get_metadata_file` instead
        fh, headers = self.service.get_metadata_file(
            trans, history_content_id=hda_id, metadata_file=metadata_name, open_file=True
        )
        trans.response.headers.update(headers)
        return fh

    def _check_dataset(self, trans, hda_id):
        # DEPRECATION: We still support unencoded ids for backward compatibility
        try:
            data = trans.sa_session.query(trans.app.model.HistoryDatasetAssociation).get(self.decode_id(hda_id))
            if data is None:
                raise ValueError(f"Invalid reference dataset id: {hda_id}.")
        except Exception:
            try:
                data = trans.sa_session.query(trans.app.model.HistoryDatasetAssociation).get(int(hda_id))
            except Exception:
                data = None
        if not data:
            raise web.httpexceptions.HTTPNotFound(f"Invalid reference dataset id: {str(hda_id)}.")
        if not self._can_access_dataset(trans, data):
            return trans.show_error_message("You are not allowed to access this dataset")
        if data.purged or data.dataset.purged:
            return trans.show_error_message("The dataset you are attempting to view has been purged.")
        elif data.deleted and not (trans.user_is_admin or (data.history and trans.get_user() == data.user)):
            return trans.show_error_message("The dataset you are attempting to view has been deleted.")
        elif data.state == Dataset.states.UPLOAD:
            return trans.show_error_message(
                "Please wait until this dataset finishes uploading before attempting to view it."
            )
        elif data.state == Dataset.states.DISCARDED:
            return trans.show_error_message("The dataset you are attempting to view has been discarded.")
        elif data.state == Dataset.states.DEFERRED:
            return trans.show_error_message(
                "The dataset you are attempting to view has deferred data. You can only use this dataset as input for jobs."
            )
        elif data.state == Dataset.states.PAUSED:
            return trans.show_error_message(
                "The dataset you are attempting to view is in paused state. One of the inputs for the job that creates this dataset has failed."
            )
        return data

    @web.expose
    def display(
        self, trans, dataset_id=None, preview=False, filename=None, to_ext=None, offset=None, ck_size=None, **kwd
    ):
        data = self._check_dataset(trans, dataset_id)
        if not isinstance(data, trans.app.model.DatasetInstance):
            return data
        if "hdca" in kwd:
            raise RequestParameterInvalidException("Invalid request parameter 'hdca' encountered.")
        hdca_id = kwd.get("hdca_id", None)
        if hdca_id:
            hdca = self.app.dataset_collection_manager.get_dataset_collection_instance(trans, "history", hdca_id)
            del kwd["hdca_id"]
            kwd["hdca"] = hdca
        # Ensure offset is an integer before passing through to datatypes.
        if offset:
            offset = int(offset)
        # Ensure ck_size is an integer before passing through to datatypes.
        if ck_size:
            ck_size = int(ck_size)
        kwd.pop("dataset", None)
        # `dataset` in kwd would interfere with positional dataset argument of `display_data` method.
        display_data, headers = data.datatype.display_data(
            trans, data, preview, filename, to_ext, offset=offset, ck_size=ck_size, **kwd
        )
        if isinstance(display_data, ZipstreamWrapper):
            trans.response.headers.update(headers)
            return display_data.response()
        trans.response.headers.update(headers)
        return display_data

    @web.legacy_expose_api_anonymous
    def get_edit(self, trans, dataset_id=None, **kwd):
        """Produces the input definitions available to modify dataset attributes"""
        status = None
        data, message = self._get_dataset_for_edit(trans, dataset_id)
        if message:
            return message

        if self._can_access_dataset(trans, data):
            if data.state == trans.model.Dataset.states.UPLOAD:
                return self.message_exception(
                    trans, "Please wait until this dataset finishes uploading before attempting to edit its metadata."
                )
            # let's not overwrite the imported datatypes module with the variable datatypes?
            # the built-in 'id' is overwritten in lots of places as well
            ldatatypes = [
                (dtype_name, dtype_name)
                for dtype_name, dtype_value in trans.app.datatypes_registry.datatypes_by_extension.items()
                if dtype_value.is_datatype_change_allowed()
            ]
            ldatatypes.sort()
            all_roles = [
                (r.name, trans.security.encode_id(r.id))
                for r in trans.app.security_agent.get_legitimate_roles(trans, data.dataset, "root")
            ]
            data_metadata = [(name, spec) for name, spec in data.metadata.spec.items()]
            converters_collection = [(key, value.name) for key, value in data.get_converter_types().items()]
            can_manage_dataset = trans.app.security_agent.can_manage_dataset(
                trans.get_current_user_roles(), data.dataset
            )
            # attribute editing
            attribute_inputs = [
                {"name": "name", "type": "text", "label": "Name", "value": data.get_display_name()},
                {"name": "info", "type": "text", "area": True, "label": "Info", "value": data.info},
                {
                    "name": "annotation",
                    "type": "text",
                    "area": True,
                    "label": "Annotation",
                    "optional": True,
                    "value": self.get_item_annotation_str(trans.sa_session, trans.user, data),
                    "help": "Add an annotation or notes to a dataset; annotations are available when a history is viewed.",
                },
            ]
            for name, spec in data_metadata:
                if spec.visible:
                    attributes = data.metadata.get_metadata_parameter(name, trans=trans)
                    if type(attributes) is form_builder.SelectField:
                        attribute_inputs.append(
                            {
                                "type": "select",
                                "multiple": attributes.multiple,
                                "optional": spec.get("optional"),
                                "name": name,
                                "label": spec.desc,
                                "options": attributes.options,
                                "value": attributes.value,
                            }
                        )
                    elif type(attributes) is form_builder.TextField:
                        attribute_inputs.append(
                            {
                                "type": "text",
                                "name": name,
                                "label": spec.desc,
                                "value": attributes.value,
                                "readonly": spec.get("readonly"),
                            }
                        )
            if data.missing_meta():
                message = 'Required metadata values are missing. Some of these values may not be editable by the user. Selecting "Auto-detect" will attempt to fix these values.'
                status = "warning"
            metadata_disable = data.state not in [
                trans.model.Dataset.states.OK,
                trans.model.Dataset.states.FAILED_METADATA,
            ]
            # datatype conversion
            conversion_options = [
                (f"{convert_id} (using '{convert_name}')", convert_id)
                for convert_id, convert_name in converters_collection
            ]
            conversion_disable = len(conversion_options) == 0
            conversion_inputs = [
                {
                    "type": "select",
                    "name": "target_type",
                    "label": "Target datatype",
                    "help": "This will create a new dataset with the contents of this dataset converted to a new format.",
                    "options": conversion_options,
                }
            ]
            # datatype changing
            datatype_options = [(ext_name, ext_id) for ext_id, ext_name in ldatatypes]
            datatype_disable = len(datatype_options) == 0
            datatype_input_default_value = None
            current_datatype = trans.app.datatypes_registry.datatypes_by_extension.get(data.ext)
            if current_datatype and current_datatype.is_datatype_change_allowed():
                datatype_input_default_value = data.ext
            datatype_inputs = [
                {
                    "type": "select",
                    "name": "datatype",
                    "label": "New Type",
                    "options": datatype_options,
                    "value": datatype_input_default_value,
                    "help": "This will change the datatype of the existing dataset but not modify its contents. Use this if Galaxy has incorrectly guessed the type of your dataset.",
                }
            ]
            # permissions
            permission_disable = True
            permission_inputs = list()
            if trans.user:
                if not data.dataset.shareable:
                    permission_message = "The dataset is stored on private storage to you and cannot be shared."
                    permission_inputs.append(
                        {"name": "not_shareable", "type": "hidden", "label": permission_message, "readonly": True}
                    )
                elif data.dataset.actions:
                    in_roles = {}
                    for action, roles in trans.app.security_agent.get_permissions(data.dataset).items():
                        in_roles[action.action] = [trans.security.encode_id(role.id) for role in roles]
                    for index, action in trans.app.model.Dataset.permitted_actions.items():
                        if action == trans.app.security_agent.permitted_actions.DATASET_ACCESS:
                            help_text = f"{action.description}<br/>NOTE: Users must have every role associated with this dataset in order to access it."
                        else:
                            help_text = action.description
                        permission_inputs.append(
                            {
                                "type": "select",
                                "multiple": True,
                                "optional": True,
                                "name": index,
                                "label": action.action,
                                "help": help_text,
                                "options": all_roles,
                                "value": in_roles.get(action.action),
                                "readonly": not can_manage_dataset,
                            }
                        )
                    permission_disable = not can_manage_dataset
                else:
                    permission_inputs.append(
                        {
                            "name": "access_public",
                            "type": "hidden",
                            "label": "This dataset is accessible by everyone (it is public).",
                            "readonly": True,
                        }
                    )
            else:
                permission_inputs.append(
                    {
                        "name": "no_access",
                        "type": "hidden",
                        "label": "Permissions not available (not logged in).",
                        "readonly": True,
                    }
                )
            return {
                "display_name": data.get_display_name(),
                "message": message,
                "status": status,
                "dataset_id": dataset_id,
                "metadata_disable": metadata_disable,
                "attribute_inputs": attribute_inputs,
                "conversion_inputs": conversion_inputs,
                "conversion_disable": conversion_disable,
                "datatype_inputs": datatype_inputs,
                "datatype_disable": datatype_disable,
                "permission_inputs": permission_inputs,
                "permission_disable": permission_disable,
            }
        else:
            return self.message_exception(
                trans, "You do not have permission to edit this dataset's ( id: %s ) information." % str(dataset_id)
            )

    @web.expose_api_anonymous
    def set_edit(self, trans, payload=None, **kwd):
        """Allows user to modify parameters of an HDA."""
        status = "success"
        operation = payload.get("operation")
        dataset_id = payload.get("dataset_id")
        data, message = self._get_dataset_for_edit(trans, dataset_id)
        if message:
            return message

        if operation == "attributes":
            # The user clicked the Save button on the 'Edit Attributes' form
            data.name = payload.get("name")
            data.info = payload.get("info")
            if data.ok_to_edit_metadata():
                # The following for loop will save all metadata_spec items
                for name, spec in data.datatype.metadata_spec.items():
                    if not spec.get("readonly"):
                        setattr(data.metadata, name, spec.unwrap(payload.get(name) or None))
                data.datatype.after_setting_metadata(data)
                # Sanitize annotation before adding it.
                if payload.get("annotation"):
                    annotation = sanitize_html(payload.get("annotation"))
                    self.add_item_annotation(trans.sa_session, trans.get_user(), data, annotation)
                # if setting metadata previously failed and all required elements have now been set, clear the failed state.
                if data._state == trans.model.Dataset.states.FAILED_METADATA and not data.missing_meta():
                    data._state = None
                message = f"Attributes updated. {message}" if message else "Attributes updated."
            else:
                message = "Attributes updated, but metadata could not be changed because this dataset is currently being used as input or output. You must cancel or wait for these jobs to complete before changing metadata."
                status = "warning"
            with transaction(trans.sa_session):
                trans.sa_session.commit()
        elif operation == "datatype":
            # The user clicked the Save button on the 'Change data type' form
            datatype = payload.get("datatype")
            self.hda_deserializer.deserialize(data, {"datatype": datatype}, trans=trans)
            message = f"Changed the type to {datatype}."
        elif operation == "datatype_detect":
            # The user clicked the 'Detect datatype' button on the 'Change data type' form
            if data.datatype.is_datatype_change_allowed():
                # prevent modifying datatype when dataset is queued or running as input/output
                if not data.ok_to_edit_metadata():
                    return self.message_exception(
                        trans,
                        "This dataset is currently being used as input or output.  You cannot change datatype until the jobs have completed or you have canceled them.",
                    )
                else:
                    path = data.dataset.file_name
                    datatype = guess_ext(path, trans.app.datatypes_registry.sniff_order)
                    trans.app.datatypes_registry.change_datatype(data, datatype)
                    with transaction(trans.sa_session):
                        trans.sa_session.commit()
                    job, *_ = trans.app.datatypes_registry.set_external_metadata_tool.tool_action.execute(
                        trans.app.datatypes_registry.set_external_metadata_tool,
                        trans,
                        incoming={"input1": data},
                        overwrite=False,
                    )  # overwrite is False as per existing behavior
                    trans.app.job_manager.enqueue(job, tool=trans.app.datatypes_registry.set_external_metadata_tool)
                    message = f"Detection was finished and changed the datatype to {datatype}."
            else:
                return self.message_exception(trans, f'Changing datatype "{data.extension}" is not allowed.')
        elif operation == "autodetect":
            # The user clicked the Auto-detect button on the 'Edit Attributes' form
            self.hda_manager.set_metadata(trans, data, overwrite=True)
            message = "Auto-detect operation successfully submitted."
        elif operation == "conversion":
            target_type = payload.get("target_type")
            if target_type:
                try:
                    message = data.datatype.convert_dataset(trans, data, target_type)
                except DatatypeConverterNotFoundException as e:
                    return self.message_exception(trans, str(e))
        elif operation == "permission":
            # Adapt form request to API - style.
            payload_permissions = {}
            for key, value in {"DATASET_MANAGE_PERMISSIONS": "manage_ids", "DATASET_ACCESS": "access_ids"}.items():
                role_ids = util.listify(payload.get(key))
                decoded_role_ids = list(map(self.decode_id, role_ids))
                payload_permissions[f"{value}[]"] = decoded_role_ids

            self.hda_manager.update_permissions(
                trans,
                data,
                action="set_permissions",
                **payload_permissions,
            )
            message = "Your changes completed successfully."
        else:
            return self.message_exception(trans, f"Invalid operation identifier ({operation}).")
        return {"status": status, "message": sanitize_text(message)}

    def _get_dataset_for_edit(self, trans, dataset_id):
        if dataset_id is not None:
            id = self.decode_id(dataset_id)
            data = trans.sa_session.query(self.app.model.HistoryDatasetAssociation).get(id)
        else:
            trans.log_event("dataset_id is None, cannot load a dataset to edit.")
            return None, self.message_exception(trans, "You must provide a dataset id to edit attributes.")
        if data is None:
            trans.log_event(f"Problem retrieving dataset id ({dataset_id}).")
            return None, self.message_exception(trans, "The dataset id is invalid.")
        if dataset_id is not None and data.user and data.user != trans.user:
            trans.log_event(f"User attempted to edit a dataset they do not own (encoded: {dataset_id}, decoded: {id}).")
            return None, self.message_exception(trans, "The dataset id is invalid.")
        if data.history.user and not data.dataset.has_manage_permissions_roles(trans.app.security_agent):
            # Permission setting related to DATASET_MANAGE_PERMISSIONS was broken for a period of time,
            # so it is possible that some Datasets have no roles associated with the DATASET_MANAGE_PERMISSIONS
            # permission.  In this case, we'll reset this permission to the hda user's private role.
            manage_permissions_action = trans.app.security_agent.get_action(
                trans.app.security_agent.permitted_actions.DATASET_MANAGE_PERMISSIONS.action
            )
            permissions = {manage_permissions_action: [trans.app.security_agent.get_private_user_role(data.user)]}
            trans.app.security_agent.set_dataset_permission(data.dataset, permissions)
        return data, None

    @web.expose
    def imp(self, trans, dataset_id=None, **kwd):
        """Import another user's dataset via a shared URL; dataset is added to user's current history."""
        # Set referer message.
        referer = trans.request.referer
        if referer and not referer.startswith(f"{trans.request.application_url}{url_for('/login')}"):
            referer_message = f"<a href='{escape(referer)}'>return to the previous page</a>"
        else:
            referer_message = f"<a href='{url_for('/')}'>go to Galaxy's start page</a>"
        # Error checking.
        if not dataset_id:
            return trans.show_error_message(
                f"You must specify a dataset to import. You can {referer_message}.", use_panels=True
            )
        # Do import.
        cur_history = trans.get_history(create=True)
        status, message = self._copy_datasets(trans, [dataset_id], [cur_history], imported=True)
        message = (
            f"Dataset imported. <br>You can <a href='{url_for('/')}'>start using the dataset</a> or {referer_message}."
        )
        return trans.show_message(message, type=status, use_panels=True)

    @web.expose
    def display_by_username_and_slug(self, trans, username, slug, filename=None, preview=True, **kwargs):
        """Display dataset by username and slug; because datasets do not yet have slugs, the slug is the dataset's id."""
        dataset = self._check_dataset(trans, slug)
        if not isinstance(dataset, trans.app.model.DatasetInstance):
            return dataset
        # Filename used for composite types.
        if filename:
            return self.display(trans, dataset_id=slug, filename=filename)

        truncated, dataset_data = self.hda_manager.text_data(dataset, preview)
        dataset.annotation = self.get_item_annotation_str(trans.sa_session, dataset.user, dataset)

        # If dataset is chunkable, get first chunk.
        first_chunk = None
        if dataset.datatype.CHUNKABLE:
            first_chunk = dataset.datatype.get_chunk(trans, dataset, 0)

        # If data is binary or an image, stream without template; otherwise, use display template.
        # TODO: figure out a way to display images in display template.
        if (
            isinstance(dataset.datatype, datatypes.binary.Binary)
            or isinstance(dataset.datatype, datatypes.images.Image)
            or isinstance(dataset.datatype, datatypes.text.Html)
        ):
            trans.response.set_content_type(dataset.get_mime())
            return open(dataset.file_name, "rb")
        else:
            return trans.fill_template_mako(
                "/dataset/display.mako",
                item=dataset,
                item_data=dataset_data,
                truncated=truncated,
                first_chunk=first_chunk,
            )

    @web.expose
    def annotate_async(self, trans, id, new_annotation=None, **kwargs):
        # TODO:?? why is this an access check only?
        decoded_id = self.decode_id(id)
        dataset = self.hda_manager.get_accessible(decoded_id, trans.user)
        dataset = self.hda_manager.error_if_uploading(dataset)
        if not dataset:
            web.httpexceptions.HTTPNotFound()
        if dataset and new_annotation:
            # Sanitize annotation before adding it.
            new_annotation = sanitize_html(new_annotation)
            self.add_item_annotation(trans.sa_session, trans.get_user(), dataset, new_annotation)
            with transaction(trans.sa_session):
                trans.sa_session.commit()
            return new_annotation

    @web.expose
    def get_annotation_async(self, trans, id):
        decoded_id = self.decode_id(id)
        dataset = self.hda_manager.get_accessible(decoded_id, trans.user)
        dataset = self.hda_manager.error_if_uploading(dataset)
        if not dataset:
            web.httpexceptions.HTTPNotFound()
        annotation = self.get_item_annotation_str(trans.sa_session, trans.user, dataset)
        if annotation and isinstance(annotation, str):
            annotation = annotation.encode("ascii", "replace")  # paste needs ascii here
        return annotation

    @web.expose
    def display_at(self, trans, dataset_id, filename=None, **kwd):
        """Sets up a dataset permissions so it is viewable at an external site"""
        if not trans.app.config.enable_old_display_applications:
            return trans.show_error_message(
                "This method of accessing external display applications has been disabled by a Galaxy administrator."
            )
        site = filename
        data = trans.sa_session.query(trans.app.model.HistoryDatasetAssociation).get(dataset_id)
        if not data:
            raise paste.httpexceptions.HTTPRequestRangeNotSatisfiable(
                f"Invalid reference dataset id: {str(dataset_id)}."
            )
        if "display_url" not in kwd or "redirect_url" not in kwd:
            return trans.show_error_message(
                'Invalid parameters specified for "display at" link, please contact a Galaxy administrator'
            )
        try:
            redirect_url = kwd["redirect_url"] % quote_plus(kwd["display_url"])
        except Exception:
            redirect_url = kwd["redirect_url"]  # not all will need custom text
        if trans.app.security_agent.dataset_is_public(data.dataset):
            return trans.response.send_redirect(redirect_url)  # anon access already permitted by rbac
        if self._can_access_dataset(trans, data):
            trans.app.host_security_agent.set_dataset_permissions(data, trans.user, site)
            return trans.response.send_redirect(redirect_url)
        else:
            return trans.show_error_message(
                "You are not allowed to view this dataset at external sites.  Please contact your Galaxy administrator to acquire management permissions for this dataset."
            )

    @web.expose
    @web.do_not_cache
    def display_application(
        self,
        trans,
        dataset_id=None,
        user_id=None,
        app_name=None,
        link_name=None,
        app_action=None,
        action_param=None,
        action_param_extra=None,
        **kwds,
    ):
        """Access to external display applications"""
        if None in [app_name, link_name]:
            return trans.show_error_message("A display application name and link name must be provided.")
        app_name = unquote_plus(app_name)
        link_name = unquote_plus(link_name)
        # Build list of parameters to pass in to display application logic (app_kwds)
        app_kwds = {}
        for name, value in dict(kwds).items():  # clone kwds because we remove stuff as we go.
            if name.startswith("app_"):
                app_kwds[name[len("app_") :]] = value
                del kwds[name]
        if kwds:
            log.debug(f"Unexpected Keywords passed to display_application: {kwds}")  # route memory?
        # decode ids
        data, user = decode_dataset_user(trans, dataset_id, user_id)
        if not data:
            raise paste.httpexceptions.HTTPRequestRangeNotSatisfiable(
                f"Invalid reference dataset id: {str(dataset_id)}."
            )
        if user is None:
            user = trans.user
        if user:
            user_roles = user.all_roles()
        else:
            user_roles = []
        # Decode application name and link name
        if self._can_access_dataset(trans, data, additional_roles=user_roles):
            msg = []
            preparable_steps = []
            refresh = False
            display_app = trans.app.datatypes_registry.display_applications.get(app_name)
            if not display_app:
                log.debug("Unknown display application has been requested: %s", app_name)
                return paste.httpexceptions.HTTPNotFound(
                    f"The requested display application ({app_name}) is not available."
                )
            dataset_hash, user_hash = encode_dataset_user(trans, data, user)
            try:
                display_link = display_app.get_link(link_name, data, dataset_hash, user_hash, trans, app_kwds)
            except Exception as e:
                log.debug("Error generating display_link: %s", e)
                # User can sometimes recover from, e.g. conversion errors by fixing input metadata, so use conflict
                return paste.httpexceptions.HTTPConflict(f"Error generating display_link: {e}")
            if not display_link:
                log.debug("Unknown display link has been requested: %s", link_name)
                return paste.httpexceptions.HTTPNotFound(f"Unknown display link has been requested: {link_name}")
            if data.state == data.states.ERROR:
                msg.append(
                    (
                        "This dataset is in an error state, you cannot view it at an external display application.",
                        "error",
                    )
                )
            elif data.deleted:
                msg.append(
                    ("This dataset has been deleted, you cannot view it at an external display application.", "error")
                )
            elif data.state != data.states.OK:
                msg.append(
                    (
                        "You must wait for this dataset to be created before you can view it at an external display application.",
                        "info",
                    )
                )
                refresh = True
            else:
                # We have permissions, dataset is not deleted and is in OK state, allow access
                if display_link.display_ready():
                    if app_action in ["data", "param"]:
                        assert action_param, "An action param must be provided for a data or param action"
                        # data is used for things with filenames that could be passed off to a proxy
                        # in case some display app wants all files to be in the same 'directory',
                        # data can be forced to param, but not the other way (no filename for other direction)
                        # get param name from url param name
                        try:
                            action_param = display_link.get_param_name_by_url(action_param)
                        except ValueError as e:
                            log.debug(e)
                            return paste.httpexceptions.HTTPNotFound(util.unicodify(e))
                        value = display_link.get_param_value(action_param)
                        assert value, f"An invalid parameter name was provided: {action_param}"
                        assert value.parameter.viewable, "This parameter is not viewable."
                        if value.parameter.type == "data":
                            try:
                                if action_param_extra:
                                    assert (
                                        value.parameter.allow_extra_files_access
                                    ), f"Extra file content requested ({action_param_extra}), but allow_extra_files_access is False."
                                    file_name = os.path.join(value.extra_files_path, action_param_extra)
                                else:
                                    file_name = value.file_name
                                content_length = os.path.getsize(file_name)
                                rval = open(file_name, "rb")
                            except OSError as e:
                                log.debug("Unable to access requested file in display application: %s", e)
                                return paste.httpexceptions.HTTPNotFound("This file is no longer available.")
                        else:
                            rval = str(value)
                            content_length = len(rval)
                        # Set Access-Control-Allow-Origin as specified in GEDA
                        if value.parameter.allow_cors:
                            trans.set_cors_origin()
                            trans.set_cors_allow()
                        trans.response.set_content_type(value.mime_type(action_param_extra=action_param_extra))
                        trans.response.headers["Content-Length"] = str(content_length)
                        return rval
                    elif app_action is None:
                        # redirect user to url generated by display link
                        return trans.response.send_redirect(display_link.display_url())
                    else:
                        msg.append((f"Invalid action provided: {app_action}", "error"))
                else:
                    if app_action is None:
                        if trans.history != data.history:
                            msg.append(
                                (
                                    "You must import this dataset into your current history before you can view it at the desired display application.",
                                    "error",
                                )
                            )
                        else:
                            refresh = True
                            msg.append(
                                (
                                    "Launching this display application required additional datasets to be generated, you can view the status of these jobs below. ",
                                    "info",
                                )
                            )
                            if not display_link.preparing_display():
                                display_link.prepare_display()
                            preparable_steps = display_link.get_prepare_steps()
                    else:
                        raise Exception(f"Attempted a view action ({app_action}) on a non-ready display application")
            return trans.fill_template_mako(
                "dataset/display_application/display.mako",
                msg=msg,
                display_app=display_app,
                display_link=display_link,
                refresh=refresh,
                preparable_steps=preparable_steps,
            )
        return trans.show_error_message(
            "You do not have permission to view this dataset at an external display application."
        )

    def _delete(self, trans, dataset_id):
        message = None
        status = "done"
        id = None
        try:
            id = self.decode_id(dataset_id)
            hda = self.hda_manager.get_owned(id, trans.user, current_history=trans.history)
            self.hda_manager.error_unless_mutable(hda.history)
            hda.mark_deleted()
            hda.clear_associated_files()
            trans.log_event(f"Dataset id {str(id)} marked as deleted")
            self.hda_manager.stop_creating_job(hda, flush=True)
        except Exception:
            msg = f"HDA deletion failed (encoded: {dataset_id}, decoded: {id})"
            log.exception(msg)
            trans.log_event(msg)
            message = "Dataset deletion failed"
            status = "error"
        return (message, status)

    def _undelete(self, trans, dataset_id):
        message = None
        status = "done"
        id = None
        try:
            id = self.decode_id(dataset_id)
            item = self.hda_manager.get_owned(id, trans.user, current_history=trans.history)
            self.hda_manager.error_unless_mutable(item.history)
            self.hda_manager.undelete(item)
            trans.log_event(f"Dataset id {str(id)} has been undeleted")
        except Exception:
            msg = f"HDA undeletion failed (encoded: {dataset_id}, decoded: {id})"
            log.exception(msg)
            trans.log_event(msg)
            message = "Dataset undeletion failed"
            status = "error"
        return (message, status)

    def _unhide(self, trans, dataset_id):
        try:
            id = self.decode_id(dataset_id)
            item = self.hda_manager.get_owned(id, trans.user, current_history=trans.history)
            self.hda_manager.error_unless_mutable(item.history)
            item.mark_unhidden()
            with transaction(trans.sa_session):
                trans.sa_session.commit()
            trans.log_event(f"Dataset id {str(id)} has been unhidden")
            return True
        except Exception:
            return False

    def _purge(self, trans, dataset_id):
        message = None
        status = "done"
        try:
            id = self.decode_id(dataset_id)
            user = trans.get_user()
            hda = trans.sa_session.query(self.app.model.HistoryDatasetAssociation).get(id)
            # Invalid HDA
            assert hda, "Invalid history dataset ID"

            # If the user is anonymous, make sure the HDA is owned by the current session.
            if not user:
                current_history_id = trans.galaxy_session.current_history_id
                assert hda.history.id == current_history_id, "Data does not belong to current user"
            # If the user is known, make sure the HDA is owned by the current user.
            else:
                assert hda.history.user == user, "Data does not belong to current user"

            # Ensure HDA is deleted
            hda.deleted = True
            # HDA is purgeable
            # Decrease disk usage first
            hda.purge_usage_from_quota(user, hda.dataset.quota_source_info)
            # Mark purged
            hda.purged = True
            trans.sa_session.add(hda)
            trans.log_event(f"HDA id {hda.id} has been purged")
            with transaction(trans.sa_session):
                trans.sa_session.commit()
            # Don't delete anything if there are active HDAs or any LDDAs, even if
            # the LDDAs are deleted.  Let the cleanup scripts get it in the latter
            # case.
            if hda.dataset.user_can_purge:
                try:
                    hda.dataset.full_delete()
                    trans.log_event(
                        f"Dataset id {hda.dataset.id} has been purged upon the the purge of HDA id {hda.id}"
                    )
                    trans.sa_session.add(hda.dataset)
                except Exception:
                    log.exception(f"Unable to purge dataset ({hda.dataset.id}) on purge of HDA ({hda.id}):")
            with transaction(trans.sa_session):
                trans.sa_session.commit()
        except Exception:
            msg = f"HDA purge failed (encoded: {dataset_id}, decoded: {id})"
            log.exception(msg)
            trans.log_event(msg)
            message = "Dataset removal from disk failed"
            status = "error"
        return (message, status)

    @web.expose
    def delete(self, trans, dataset_id, filename, show_deleted_on_refresh=False):
        message, status = self._delete(trans, dataset_id)
        return trans.response.send_redirect(
            web.url_for(
                controller="root",
                action="history",
                show_deleted=show_deleted_on_refresh,
                message=message,
                status=status,
            )
        )

    @web.expose
    def delete_async(self, trans, dataset_id, filename):
        message, status = self._delete(trans, dataset_id)
        if status == "done":
            return "OK"
        else:
            raise Exception(message)

    @web.expose
    def undelete(self, trans, dataset_id, filename):
        message, status = self._undelete(trans, dataset_id)
        return trans.response.send_redirect(
            web.url_for(controller="root", action="history", show_deleted=True, message=message, status=status)
        )

    @web.expose
    def undelete_async(self, trans, dataset_id, filename):
        message, status = self._undelete(trans, dataset_id)
        if status == "done":
            return "OK"
        else:
            raise Exception(message)

    @web.expose
    def unhide(self, trans, dataset_id, filename):
        if self._unhide(trans, dataset_id):
            return trans.response.send_redirect(web.url_for(controller="root", action="history", show_hidden=True))
        raise Exception("Error unhiding")

    @web.expose
    def purge(self, trans, dataset_id, filename, show_deleted_on_refresh=False):
        if trans.app.config.allow_user_dataset_purge:
            message, status = self._purge(trans, dataset_id)
        else:
            message = "Removal of datasets by users is not allowed in this Galaxy instance.  Please contact your Galaxy administrator."
            status = "error"
        return trans.response.send_redirect(
            web.url_for(
                controller="root",
                action="history",
                show_deleted=show_deleted_on_refresh,
                message=message,
                status=status,
            )
        )

    @web.expose
    def purge_async(self, trans, dataset_id, filename):
        if trans.app.config.allow_user_dataset_purge:
            message, status = self._purge(trans, dataset_id)
        else:
            message = "Removal of datasets by users is not allowed in this Galaxy instance.  Please contact your Galaxy administrator."
            status = "error"
        if status == "done":
            return "OK"
        else:
            raise Exception(message)

    @web.expose
    def copy_datasets(
        self,
        trans,
        source_history=None,
        source_content_ids="",
        target_history_id=None,
        target_history_ids="",
        new_history_name="",
        do_copy=False,
        **kwd,
    ):
        user = trans.get_user()
        if source_history is not None:
            decoded_source_history_id = self.decode_id(source_history)
            history = self.history_manager.get_owned(
                decoded_source_history_id, trans.user, current_history=trans.history
            )
            current_history = trans.get_history()
        else:
            history = current_history = trans.get_history()
        refresh_frames = []
        if source_content_ids:
            if not isinstance(source_content_ids, list):
                source_content_ids = source_content_ids.split(",")
            encoded_dataset_collection_ids = [
                s[len("dataset_collection|") :] for s in source_content_ids if s.startswith("dataset_collection|")
            ]
            encoded_dataset_ids = [s[len("dataset|") :] for s in source_content_ids if s.startswith("dataset|")]
            decoded_dataset_collection_ids = set(map(self.decode_id, encoded_dataset_collection_ids))
            decoded_dataset_ids = set(map(self.decode_id, encoded_dataset_ids))
        else:
            decoded_dataset_collection_ids = []
            decoded_dataset_ids = []
        if new_history_name:
            target_history_ids = []
        else:
            if target_history_id:
                target_history_ids = [self.decode_id(target_history_id)]
            elif target_history_ids:
                if not isinstance(target_history_ids, list):
                    target_history_ids = target_history_ids.split(",")
                target_history_ids = list({self.decode_id(h) for h in target_history_ids if h})
            else:
                target_history_ids = []
        done_msg = error_msg = ""
        new_history = None
        if do_copy:
            invalid_contents = 0
            if not (decoded_dataset_ids or decoded_dataset_collection_ids) or not (
                target_history_ids or new_history_name
            ):
                error_msg = "You must provide both source datasets and target histories. "
            else:
                if new_history_name:
                    new_history = trans.app.model.History()
                    new_history.name = new_history_name
                    new_history.user = user
                    trans.sa_session.add(new_history)
                    with transaction(trans.sa_session):
                        trans.sa_session.commit()
                    target_history_ids.append(new_history.id)
                if user:
                    target_histories = [
                        hist
                        for hist in map(trans.sa_session.query(trans.app.model.History).get, target_history_ids)
                        if hist is not None and hist.user == user
                    ]
                else:
                    target_histories = [history]
                if len(target_histories) != len(target_history_ids):
                    error_msg = (
                        error_msg
                        + "You do not have permission to add datasets to %i requested histories.  "
                        % (len(target_history_ids) - len(target_histories))
                    )
                source_contents = list(
                    map(trans.sa_session.query(trans.app.model.HistoryDatasetAssociation).get, decoded_dataset_ids)
                )
                source_contents.extend(
                    map(
                        trans.sa_session.query(trans.app.model.HistoryDatasetCollectionAssociation).get,
                        decoded_dataset_collection_ids,
                    )
                )
                source_contents.sort(key=lambda content: content.hid)
                for content in source_contents:
                    if content is None:
                        error_msg = f"{error_msg}You tried to copy a dataset that does not exist. "
                        invalid_contents += 1
                    elif content.history != history:
                        error_msg = f"{error_msg}You tried to copy a dataset which is not in your current history. "
                        invalid_contents += 1
                    else:
                        for hist in target_histories:
                            if content.history_content_type == "dataset":
                                copy = content.copy(flush=False)
                                hist.stage_addition(copy)
                            else:
                                copy = content.copy(element_destination=hist)
                            if user:
                                copy.copy_tags_from(user, content)
                        for hist in target_histories:
                            hist.add_pending_items()
                with transaction(trans.sa_session):
                    trans.sa_session.commit()
                if current_history in target_histories:
                    refresh_frames = ["history"]
                hist_names_str = ", ".join(
                    '<a href="{}" target="_top">{}</a>'.format(
                        url_for(
                            controller="history", action="switch_to_history", hist_id=trans.security.encode_id(hist.id)
                        ),
                        escape(hist.name),
                    )
                    for hist in target_histories
                )
                num_source = len(source_content_ids) - invalid_contents
                num_target = len(target_histories)
                done_msg = "%i %s copied to %i %s: %s." % (
                    num_source,
                    inflector.cond_plural(num_source, "dataset"),
                    num_target,
                    inflector.cond_plural(num_target, "history"),
                    hist_names_str,
                )
                trans.sa_session.refresh(history)
        source_contents = history.active_contents
        target_histories = [history]
        if user:
            target_histories = user.active_histories
        return trans.fill_template(
            "/dataset/copy_view.mako",
            source_history=history,
            current_history=current_history,
            source_content_ids=source_content_ids,
            target_history_id=target_history_id,
            target_history_ids=target_history_ids,
            source_contents=source_contents,
            target_histories=target_histories,
            new_history_name=new_history_name,
            done_msg=done_msg,
            error_msg=error_msg,
            refresh_frames=refresh_frames,
        )

    def _copy_datasets(self, trans, dataset_ids, target_histories, imported=False):
        """Helper method for copying datasets."""
        user = trans.get_user()
        done_msg = error_msg = ""

        invalid_datasets = 0
        if not dataset_ids or not target_histories:
            error_msg = "You must provide both source datasets and target histories."
        else:
            # User must own target histories to copy datasets to them.
            for history in target_histories:
                if user != history.user:
                    error_msg = (
                        error_msg
                        + "You do not have permission to add datasets to %i requested histories.  "
                        % (len(target_histories))
                    )
            for dataset_id in dataset_ids:
                decoded_id = self.decode_id(dataset_id)
                data = self.hda_manager.get_accessible(decoded_id, trans.user)
                data = self.hda_manager.error_if_uploading(data)

                if data is None:
                    error_msg = f"{error_msg}You tried to copy a dataset that does not exist or that you do not have access to.  "
                    invalid_datasets += 1
                else:
                    for hist in target_histories:
                        dataset_copy = data.copy()
                        if imported:
                            dataset_copy.name = f"imported: {dataset_copy.name}"
                        hist.add_dataset(dataset_copy)
            with transaction(trans.sa_session):
                trans.sa_session.commit()
            num_datasets_copied = len(dataset_ids) - invalid_datasets
            done_msg = "%i dataset%s copied to %i histor%s." % (
                num_datasets_copied,
                iff(num_datasets_copied == 1, "", "s"),
                len(target_histories),
                iff(len(target_histories) == 1, "y", "ies"),
            )
            trans.sa_session.refresh(history)

        if error_msg != "":
            status = ERROR
            message = error_msg
        else:
            status = SUCCESS
            message = done_msg
        return status, message



import logging

import tool_shed.util.shed_util_common as suc
from galaxy import (
    exceptions,
    util,
    web,
)
from galaxy.model.base import transaction
from galaxy.security.validate_user_input import (
    validate_email,
    validate_password,
    validate_publicname,
)
from galaxy.webapps.base.controller import BaseAPIController

log = logging.getLogger(__name__)


class UsersController(BaseAPIController):
    """RESTful controller for interactions with users in the Tool Shed."""

    @web.expose_api
    @web.require_admin
    def create(self, trans, payload, **kwd):
        """
                POST /api/users
                Returns a dictionary of information about the created user.

        :       param key: the current Galaxy admin user's API key

                The following parameters are included in the payload.
                :param email (required): the email address of the user
                :param password (required): the password of the user
                :param username (required): the public username of the user
        """
        # Get the information about the user to be created from the payload.
        email = payload.get("email", "")
        password = payload.get("password", "")
        username = payload.get("username", "")
        message = self.__validate(trans, email=email, password=password, confirm=password, username=username)
        if message:
            raise exceptions.RequestParameterInvalidException(message)

        # Create the user.
        user = self.__create_user(trans, email, username, password)
        user_dict = user.to_dict(view="element", value_mapper=self.__get_value_mapper(trans))
        user_dict["message"] = f"User '{str(user.username)}' has been created."
        user_dict["url"] = web.url_for(controller="users", action="show", id=trans.security.encode_id(user.id))
        return user_dict

    def __create_user(self, trans, email, username, password):
        user = trans.app.model.User(email=email)
        user.set_password_cleartext(password)
        user.username = username
        if trans.app.config.user_activation_on:
            user.active = False
        else:
            user.active = True  # Activation is off, every new user is active by default.
        trans.sa_session.add(user)
        with transaction(trans.sa_session):
            trans.sa_session.commit()
        trans.app.security_agent.create_private_user_role(user)
        return user

    def __get_value_mapper(self, trans):
        value_mapper = {"id": trans.security.encode_id}
        return value_mapper

    @web.expose_api_anonymous
    def index(self, trans, deleted=False, **kwd):
        """
        GET /api/users
        Returns a list of dictionaries that contain information about each user.
        """
        # Example URL: http://localhost:9009/api/users
        user_dicts = []
        deleted = util.asbool(deleted)
        for user in (
            trans.sa_session.query(trans.app.model.User)
            .filter(trans.app.model.User.table.c.deleted == deleted)
            .order_by(trans.app.model.User.table.c.username)
        ):
            user_dict = user.to_dict(view="collection", value_mapper=self.__get_value_mapper(trans))
            user_dict["url"] = web.url_for(controller="users", action="show", id=trans.security.encode_id(user.id))
            user_dicts.append(user_dict)
        return user_dicts

    @web.expose_api_anonymous
    def show(self, trans, id, **kwd):
        """
        GET /api/users/{encoded_user_id}
        GET /api/users/current
        Returns a dictionary of information about a user.

        :param id: the encoded id of the User object.
        """
        user = None
        # user is requesting data about themselves
        user = trans.user if id == "current" else suc.get_user(trans.app, id)
        if user is None:
            user_dict = dict(message=f"Unable to locate user record for id {str(id)}.", status="error")
            return user_dict
        user_dict = user.to_dict(view="element", value_mapper=self.__get_value_mapper(trans))
        user_dict["url"] = web.url_for(controller="users", action="show", id=trans.security.encode_id(user.id))
        return user_dict

    def __validate(self, trans, email, password, confirm, username):
        if username in ["repos"]:
            return f"The term '{username}' is a reserved word in the Tool Shed, so it cannot be used as a public user name."
        message = "\n".join(
            (
                validate_email(trans, email),
                validate_password(trans, password, confirm),
                validate_publicname(trans, username),
            )
        ).rstrip()
        return message



from typing import Optional

from galaxy import exceptions as glx_exceptions
from galaxy.managers import api_keys
from galaxy.managers.context import ProvidesUserContext
from galaxy.managers.users import UserManager
from galaxy.queue_worker import send_local_control_task
from galaxy.schema import APIKeyModel
from galaxy.security.idencoding import IdEncodingHelper
from galaxy.webapps.galaxy.services.base import (
    async_task_summary,
    ServiceBase,
)


class UsersService(ServiceBase):
    """Common interface/service logic for interactions with users in the context of the API.

    Provides the logic of the actions invoked by API controllers and uses type definitions
    and pydantic models to declare its parameters and return types.
    """

    def __init__(self, security: IdEncodingHelper, user_manager: UserManager, api_key_manager: api_keys.ApiKeyManager):
        super().__init__(security)
        self.user_manager = user_manager
        self.api_key_manager = api_key_manager

    def recalculate_disk_usage(
        self,
        trans: ProvidesUserContext,
    ):
        if trans.anonymous:
            raise glx_exceptions.AuthenticationRequired("Only registered users can recalculate disk usage.")
        if trans.app.config.enable_celery_tasks:
            from galaxy.celery.tasks import recalculate_user_disk_usage

            result = recalculate_user_disk_usage.delay(user_id=trans.user.id)
            return async_task_summary(result)
        else:
            send_local_control_task(
                trans.app,
                "recalculate_user_disk_usage",
                kwargs={
                    "user_id": trans.user.id,
                },
            )
            return None

    def get_api_key(self, trans: ProvidesUserContext, user_id: int) -> Optional[APIKeyModel]:
        """Returns the current API key or None if the user doesn't have any valid API key."""
        user = self._get_user(trans, user_id)
        api_key = self.api_key_manager.get_api_key(user)
        return APIKeyModel.construct(key=api_key.key, create_time=api_key.create_time) if api_key else None

    def get_or_create_api_key(self, trans: ProvidesUserContext, user_id: int) -> str:
        """Returns the current API key (as plain string) or creates a new one."""
        user = self._get_user(trans, user_id)
        return self.api_key_manager.get_or_create_api_key(user)

    def create_api_key(self, trans: ProvidesUserContext, user_id: int) -> APIKeyModel:
        """Creates a new API key for the given user"""
        user = self._get_user(trans, user_id)
        api_key = self.api_key_manager.create_api_key(user)
        result = APIKeyModel.construct(key=api_key.key, create_time=api_key.create_time)
        return result

    def delete_api_key(self, trans: ProvidesUserContext, user_id: int) -> None:
        """Deletes a particular API key"""
        user = self._get_user(trans, user_id)
        self.api_key_manager.delete_api_key(user)

    def _get_user(self, trans: ProvidesUserContext, user_id):
        user = trans.user
        if trans.anonymous or (user and user.id != user_id and not trans.user_is_admin):
            raise glx_exceptions.InsufficientPermissionsException("Access denied.")
        user = self.user_manager.by_id(user_id)
        return user



"""
Manager and Serializer for TS repositories.
"""
import logging

from sqlalchemy.orm.exc import (
    MultipleResultsFound,
    NoResultFound,
)

from galaxy.exceptions import (
    InconsistentDatabase,
    InternalServerError,
    RequestParameterInvalidException,
)

log = logging.getLogger(__name__)


# =============================================================================
class RepoManager:
    """
    Interface/service object for interacting with TS repositories.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get(self, trans, decoded_repo_id):
        """
        Get the repo from the DB.

        :param  decoded_repo_id:       decoded repo id
        :type   decoded_repo_id:       int

        :returns:   the requested repo
        :rtype:     tool_shed.webapp.model.Repository
        """
        try:
            repo = (
                trans.sa_session.query(trans.app.model.Repository)
                .filter(trans.app.model.Repository.table.c.id == decoded_repo_id)
                .one()
            )
        except MultipleResultsFound:
            raise InconsistentDatabase("Multiple repositories found with the same id.")
        except NoResultFound:
            raise RequestParameterInvalidException("No repository found with the id provided.")
        except Exception:
            raise InternalServerError("Error loading from the database.")
        return repo

    def list_by_owner(self, trans, user_id):
        """
        Return a list of of repositories owned by a given TS user from the DB.

        :returns: query that will emit repositories owned by given user
        :rtype:   sqlalchemy query
        """
        query = trans.sa_session.query(trans.app.model.Repository).filter(
            trans.app.model.Repository.table.c.user_id == user_id
        )
        return query

    def create(self, trans, name, description=""):
        """
        Create a new group.
        """

    def update(self, trans, group, name=None, description=None):
        """
        Update the given group
        """

    def delete(self, trans, group, undelete=False):
        """
        Mark given group deleted/undeleted based on the flag.
        """



"""
API operations on a jobs.

.. seealso:: :class:`galaxy.model.Jobs`
"""

import logging
from datetime import (
    date,
    datetime,
)
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
)

from fastapi import (
    Depends,
    Query,
)

from galaxy import (
    exceptions,
    model,
)
from galaxy.managers import hdas
from galaxy.managers.context import (
    ProvidesHistoryContext,
    ProvidesUserContext,
)
from galaxy.managers.jobs import (
    JobLock,
    JobManager,
    JobSearch,
    summarize_destination_params,
    summarize_job_metrics,
    summarize_job_parameters,
)
from galaxy.schema.fields import DecodedDatabaseIdField
from galaxy.schema.schema import JobIndexSortByEnum
from galaxy.schema.types import OffsetNaiveDatetime
from galaxy.web import (
    expose_api,
    expose_api_anonymous,
    require_admin,
)
from galaxy.webapps.base.controller import UsesVisualizationMixin
from galaxy.webapps.galaxy.api import (
    BaseGalaxyAPIController,
    depends,
    DependsOnTrans,
    IndexQueryTag,
    Router,
    search_query_param,
)
from galaxy.webapps.galaxy.api.common import query_parameter_as_list
from galaxy.webapps.galaxy.services.jobs import (
    JobIndexPayload,
    JobIndexViewEnum,
    JobsService,
)
from galaxy.work.context import WorkRequestContext

log = logging.getLogger(__name__)

router = Router(tags=["jobs"])


StateQueryParam = Query(
    default=None,
    alias="state",
    title="States",
    description="A list or comma-separated list of states to filter job query on. If unspecified, jobs of any state may be returned.",
)

UserDetailsQueryParam: bool = Query(
    default=False,
    title="Include user details",
    description="If true, and requester is an admin, will return external job id and user email. This is only available to admins.",
)

UserIdQueryParam: Optional[DecodedDatabaseIdField] = Query(
    default=None,
    title="User ID",
    description="an encoded user id to restrict query to, must be own id if not admin user",
)

ViewQueryParam: JobIndexViewEnum = Query(
    default="collection",
    title="View",
    description="Determines columns to return. Defaults to 'collection'.",
)


ToolIdQueryParam = Query(
    default=None,
    alias="tool_id",
    title="Tool ID(s)",
    description="Limit listing of jobs to those that match one of the included tool_ids. If none, all are returned",
)


ToolIdLikeQueryParam = Query(
    default=None,
    alias="tool_id_like",
    title="Tool ID Pattern(s)",
    description="Limit listing of jobs to those that match one of the included tool ID sql-like patterns. If none, all are returned",
)

DateRangeMinQueryParam: Optional[Union[OffsetNaiveDatetime, date]] = Query(
    default=None,
    title="Date Range Minimum",
    description="Limit listing of jobs to those that are updated after specified date (e.g. '2014-01-01')",
)

DateRangeMaxQueryParam: Optional[Union[OffsetNaiveDatetime, date]] = Query(
    default=None,
    title="Date Range Maximum",
    description="Limit listing of jobs to those that are updated before specified date (e.g. '2014-01-01')",
)

HistoryIdQueryParam: Optional[DecodedDatabaseIdField] = Query(
    default=None,
    title="History ID",
    description="Limit listing of jobs to those that match the history_id. If none, jobs from any history may be returned.",
)

WorkflowIdQueryParam: Optional[DecodedDatabaseIdField] = Query(
    default=None,
    title="Workflow ID",
    description="Limit listing of jobs to those that match the specified workflow ID. If none, jobs from any workflow (or from no workflows) may be returned.",
)

InvocationIdQueryParam: Optional[DecodedDatabaseIdField] = Query(
    default=None,
    title="Invocation ID",
    description="Limit listing of jobs to those that match the specified workflow invocation ID. If none, jobs from any workflow invocation (or from no workflows) may be returned.",
)

SortByQueryParam: JobIndexSortByEnum = Query(
    default=JobIndexSortByEnum.update_time,
    title="Sort By",
    description="Sort results by specified field.",
)

LimitQueryParam: int = Query(default=500, title="Limit", description="Maximum number of jobs to return.")

OffsetQueryParam: int = Query(
    default=0,
    title="Offset",
    description="Return jobs starting from this specified position. For example, if ``limit`` is set to 100 and ``offset`` to 200, jobs 200-299 will be returned.",
)

query_tags = [
    IndexQueryTag("user", "The user email of the user that executed the Job.", "u"),
    IndexQueryTag("tool_id", "The tool ID corresponding to the job.", "t"),
    IndexQueryTag("runner", "The job runner name used to execte the job.", "r", admin_only=True),
    IndexQueryTag("handler", "The job handler name used to execute the job.", "h", admin_only=True),
]

SearchQueryParam: Optional[str] = search_query_param(
    model_name="Job",
    tags=query_tags,
    free_text_fields=["user", "tool", "handler", "runner"],
)


@router.cbv
class FastAPIJobs:
    service: JobsService = depends(JobsService)

    @router.get("/api/jobs/{id}")
    def show(
        self,
        id: DecodedDatabaseIdField,
        trans: ProvidesUserContext = DependsOnTrans,
        full: Optional[bool] = False,
    ) -> Dict[str, Any]:
        """
        Return dictionary containing description of job data

        Parameters
        - id: ID of job to return
        - full: Return extra information ?
        """
        return self.service.show(trans, id, bool(full))

    @router.get("/api/jobs")
    def index(
        self,
        trans: ProvidesUserContext = DependsOnTrans,
        states: Optional[List[str]] = Depends(query_parameter_as_list(StateQueryParam)),
        user_details: bool = UserDetailsQueryParam,
        user_id: Optional[DecodedDatabaseIdField] = UserIdQueryParam,
        view: JobIndexViewEnum = ViewQueryParam,
        tool_ids: Optional[List[str]] = Depends(query_parameter_as_list(ToolIdQueryParam)),
        tool_ids_like: Optional[List[str]] = Depends(query_parameter_as_list(ToolIdLikeQueryParam)),
        date_range_min: Optional[Union[datetime, date]] = DateRangeMinQueryParam,
        date_range_max: Optional[Union[datetime, date]] = DateRangeMaxQueryParam,
        history_id: Optional[DecodedDatabaseIdField] = HistoryIdQueryParam,
        workflow_id: Optional[DecodedDatabaseIdField] = WorkflowIdQueryParam,
        invocation_id: Optional[DecodedDatabaseIdField] = InvocationIdQueryParam,
        order_by: JobIndexSortByEnum = SortByQueryParam,
        search: Optional[str] = SearchQueryParam,
        limit: int = LimitQueryParam,
        offset: int = OffsetQueryParam,
    ) -> List[Dict[str, Any]]:
        payload = JobIndexPayload.construct(
            states=states,
            user_details=user_details,
            user_id=user_id,
            view=view,
            tool_ids=tool_ids,
            tool_ids_like=tool_ids_like,
            date_range_min=date_range_min,
            date_range_max=date_range_max,
            history_id=history_id,
            workflow_id=workflow_id,
            invocation_id=invocation_id,
            order_by=order_by,
            search=search,
            limit=limit,
            offset=offset,
        )
        return self.service.index(trans, payload)


class JobController(BaseGalaxyAPIController, UsesVisualizationMixin):
    job_manager = depends(JobManager)
    job_search = depends(JobSearch)
    hda_manager = depends(hdas.HDAManager)

    @expose_api
    def common_problems(self, trans: ProvidesUserContext, id, **kwd):
        """
        * GET /api/jobs/{id}/common_problems
            check inputs and job for common potential problems to aid in error reporting
        """
        job = self.__get_job(trans, id)
        seen_ids = set()
        has_empty_inputs = False
        has_duplicate_inputs = False
        for job_input_assoc in job.input_datasets:
            input_dataset_instance = job_input_assoc.dataset
            if input_dataset_instance is None:
                continue
            if input_dataset_instance.get_total_size() == 0:
                has_empty_inputs = True
            input_instance_id = input_dataset_instance.id
            if input_instance_id in seen_ids:
                has_duplicate_inputs = True
            else:
                seen_ids.add(input_instance_id)
        # TODO: check percent of failing jobs around a window on job.update_time for handler - report if high.
        # TODO: check percent of failing jobs around a window on job.update_time for destination_id - report if high.
        # TODO: sniff inputs (add flag to allow checking files?)
        return {"has_empty_inputs": has_empty_inputs, "has_duplicate_inputs": has_duplicate_inputs}

    @expose_api
    def inputs(self, trans: ProvidesUserContext, id, **kwd) -> List[dict]:
        """
        GET /api/jobs/{id}/inputs

        returns input datasets created by job

        :type   id: string
        :param  id: Encoded job id

        :rtype:     list of dicts
        :returns:   list of dictionaries containing input dataset associations
        """
        job = self.__get_job(trans, id)
        return self.__dictify_associations(trans, job.input_datasets, job.input_library_datasets)

    @expose_api
    def outputs(self, trans: ProvidesUserContext, id, **kwd) -> List[dict]:
        """
        outputs( trans, id )
        * GET /api/jobs/{id}/outputs
            returns output datasets created by job

        :type   id: string
        :param  id: Encoded job id

        :rtype:     list of dicts
        :returns:   list of dictionaries containing output dataset associations
        """
        job = self.__get_job(trans, id)
        return self.__dictify_associations(trans, job.output_datasets, job.output_library_datasets)

    @expose_api
    def delete(self, trans: ProvidesUserContext, id, **kwd):
        """
        delete( trans, id )
        * Delete /api/jobs/{id}
            cancels specified job

        :type   id: string
        :param  id: Encoded job id
        :type   message: string
        :param  message: Stop message.
        """
        payload = kwd.get("payload") or {}
        job = self.__get_job(trans, id)
        message = payload.get("message", None)
        return self.job_manager.stop(job, message=message)

    @expose_api
    def resume(self, trans: ProvidesUserContext, id, **kwd) -> List[dict]:
        """
        * PUT /api/jobs/{id}/resume
            Resumes a paused job

        :type   id: string
        :param  id: Encoded job id

        :rtype:     list of dicts
        :returns:   list of dictionaries containing output dataset associations
        """
        job = self.__get_job(trans, id)
        if not job:
            raise exceptions.ObjectNotFound("Could not access job with the given id")
        if job.state == job.states.PAUSED:
            job.resume()
        else:
            exceptions.RequestParameterInvalidException(f"Job with id '{job.tool_id}' is not paused")
        return self.__dictify_associations(trans, job.output_datasets, job.output_library_datasets)

    @expose_api_anonymous
    def metrics(self, trans: ProvidesUserContext, **kwd):
        """
        * GET /api/jobs/{job_id}/metrics
        * GET /api/datasets/{dataset_id}/metrics
            Return job metrics for specified job. Job accessibility checks are slightly
            different than dataset checks, so both methods are available.

        :type   job_id: string
        :param  job_id: Encoded job id

        :type   dataset_id: string
        :param  dataset_id: Encoded HDA or LDDA id

        :type   hda_ldda: string
        :param  hda_ldda: hda if dataset_id is an HDA id (default), ldda if
                          it is an ldda id.

        :rtype:     list
        :returns:   list containing job metrics
        """
        job = self.__get_job(trans, **kwd)
        return summarize_job_metrics(trans, job)

    @require_admin
    @expose_api
    def destination_params(self, trans: ProvidesUserContext, **kwd):
        """
        * GET /api/jobs/{job_id}/destination_params
            Return destination parameters for specified job.

        :type   job_id: string
        :param  job_id: Encoded job id

        :rtype:     list
        :returns:   list containing job destination parameters
        """
        job = self.__get_job(trans, **kwd)
        return summarize_destination_params(trans, job)

    @expose_api_anonymous
    def parameters_display(self, trans: ProvidesUserContext, **kwd):
        """
        * GET /api/jobs/{job_id}/parameters_display
        * GET /api/datasets/{dataset_id}/parameters_display

            Resolve parameters as a list for nested display. More client logic
            here than is ideal but it is hard to reason about tool parameter
            types on the client relative to the server. Job accessibility checks
            are slightly different than dataset checks, so both methods are
            available.

            This API endpoint is unstable and tied heavily to Galaxy's JS client code,
            this endpoint will change frequently.

        :type   job_id: string
        :param  job_id: Encoded job id

        :type   dataset_id: string
        :param  dataset_id: Encoded HDA or LDDA id

        :type   hda_ldda: string
        :param  hda_ldda: hda if dataset_id is an HDA id (default), ldda if
                          it is an ldda id.

        :rtype:     list
        :returns:   job parameters for for display
        """
        job = self.__get_job(trans, **kwd)
        return summarize_job_parameters(trans, job)

    @expose_api_anonymous
    def build_for_rerun(self, trans: ProvidesHistoryContext, id, **kwd):
        """
        * GET /api/jobs/{id}/build_for_rerun
            returns a tool input/param template prepopulated with this job's
            information, suitable for rerunning or rendering parameters of the
            job.

        :type   id: string
        :param  id: Encoded job id

        :rtype:     dictionary
        :returns:   dictionary containing output dataset associations
        """

        job = self.__get_job(trans, id)
        if not job:
            raise exceptions.ObjectNotFound("Could not access job with the given id")
        tool = self.app.toolbox.get_tool(job.tool_id, kwd.get("tool_version") or job.tool_version)
        if tool is None:
            raise exceptions.ObjectNotFound("Requested tool not found")
        if not tool.is_workflow_compatible:
            raise exceptions.ConfigDoesNotAllowException(f"Tool '{job.tool_id}' cannot be rerun.")
        return tool.to_json(trans, {}, job=job)

    def __dictify_associations(self, trans, *association_lists) -> List[dict]:
        rval: List[dict] = []
        for association_list in association_lists:
            rval.extend(self.__dictify_association(trans, a) for a in association_list)
        return rval

    def __dictify_association(self, trans, job_dataset_association) -> dict:
        dataset_dict = None
        dataset = job_dataset_association.dataset
        if dataset:
            if isinstance(dataset, model.HistoryDatasetAssociation):
                dataset_dict = dict(src="hda", id=trans.security.encode_id(dataset.id))
            else:
                dataset_dict = dict(src="ldda", id=trans.security.encode_id(dataset.id))
        return dict(name=job_dataset_association.name, dataset=dataset_dict)

    def __get_job(self, trans, job_id=None, dataset_id=None, **kwd):
        if job_id is not None:
            decoded_job_id = self.decode_id(job_id)
            return self.job_manager.get_accessible_job(trans, decoded_job_id)
        else:
            hda_ldda = kwd.get("hda_ldda", "hda")
            # Following checks dataset accessible
            dataset_instance = self.get_hda_or_ldda(trans, hda_ldda=hda_ldda, dataset_id=dataset_id)
            return dataset_instance.creating_job

    @expose_api
    def create(self, trans: ProvidesUserContext, payload, **kwd):
        """See the create method in tools.py in order to submit a job."""
        raise exceptions.NotImplemented("Please POST to /api/tools instead.")

    @expose_api
    def search(self, trans: ProvidesHistoryContext, payload: dict, **kwd):
        """
        search( trans, payload )
        * POST /api/jobs/search:
            return jobs for current user

        :type   payload: dict
        :param  payload: Dictionary containing description of requested job. This is in the same format as
            a request to POST /apt/tools would take to initiate a job

        :rtype:     list
        :returns:   list of dictionaries containing summary job information of the jobs that match the requested job run

        This method is designed to scan the list of previously run jobs and find records of jobs that had
        the exact some input parameters and datasets. This can be used to minimize the amount of repeated work, and simply
        recycle the old results.
        """
        tool_id = payload.get("tool_id")
        if tool_id is None:
            raise exceptions.RequestParameterMissingException("No tool id")
        tool = trans.app.toolbox.get_tool(tool_id)
        if tool is None:
            raise exceptions.ObjectNotFound("Requested tool not found")
        if "inputs" not in payload:
            raise exceptions.RequestParameterMissingException("No inputs defined")
        inputs = payload.get("inputs", {})
        # Find files coming in as multipart file data and add to inputs.
        for k, v in payload.items():
            if k.startswith("files_") or k.startswith("__files_"):
                inputs[k] = v
        request_context = WorkRequestContext(app=trans.app, user=trans.user, history=trans.history)
        all_params, all_errors, _, _ = tool.expand_incoming(
            trans=trans, incoming=inputs, request_context=request_context
        )
        if any(all_errors):
            return []
        params_dump = [tool.params_to_strings(param, self.app, nested=True) for param in all_params]
        jobs = []
        for param_dump, param in zip(params_dump, all_params):
            job = self.job_search.by_tool_input(
                trans=trans,
                tool_id=tool_id,
                tool_version=tool.version,
                param=param,
                param_dump=param_dump,
                job_state=payload.get("state"),
            )
            if job:
                jobs.append(job)
        return [self.encode_all_ids(trans, single_job.to_dict("element"), True) for single_job in jobs]

    @expose_api_anonymous
    def error(self, trans: ProvidesUserContext, id, payload, **kwd):
        """
        error( trans, id )
        * POST /api/jobs/{id}/error
            submits a bug report via the API.

        :type   id: string
        :param  id: Encoded job id

        :rtype:     dictionary
        :returns:   dictionary containing information regarding where the error report was sent.
        """
        # Get dataset on which this error was triggered
        dataset_id = payload.get("dataset_id")
        if not dataset_id:
            raise exceptions.RequestParameterMissingException("No dataset_id")
        decoded_dataset_id = self.decode_id(dataset_id)
        dataset = self.hda_manager.get_accessible(decoded_dataset_id, trans.user)

        # Get job
        job = self.__get_job(trans, id)
        if dataset.creating_job.id != job.id:
            raise exceptions.RequestParameterInvalidException("dataset_id was not created by job_id")
        tool = trans.app.toolbox.get_tool(job.tool_id, tool_version=job.tool_version) or None
        email = payload.get("email")
        if not email and not trans.anonymous:
            email = trans.user.email
        messages = trans.app.error_reports.default_error_plugin.submit_report(
            dataset=dataset,
            job=job,
            tool=tool,
            user_submission=True,
            user=trans.user,
            email=email,
            message=payload.get("message"),
        )

        return {"messages": messages}

    @require_admin
    @expose_api
    def show_job_lock(self, trans: ProvidesUserContext, **kwd):
        """
        * GET /api/job_lock
            return boolean indicating if job lock active.
        """
        return self.job_manager.job_lock()

    @require_admin
    @expose_api
    def update_job_lock(self, trans: ProvidesUserContext, payload, **kwd):
        """
        * PUT /api/job_lock
            return boolean indicating if job lock active.
        """
        active = payload.get("active")
        job_lock = JobLock(active=active)
        return self.job_manager.update_job_lock(job_lock)



import json
import logging
import os
import tarfile
from collections import namedtuple
from io import StringIO
from time import strftime
from typing import (
    Callable,
    Dict,
)

from sqlalchemy import (
    and_,
    false,
)
from webob.compat import cgi_FieldStorage

from galaxy import (
    util,
    web,
)
from galaxy.exceptions import (
    ActionInputError,
    ConfigDoesNotAllowException,
    InsufficientPermissionsException,
    ObjectNotFound,
    RequestParameterInvalidException,
    RequestParameterMissingException,
)
from galaxy.util import checkers
from galaxy.web import (
    expose_api,
    expose_api_anonymous_and_sessionless,
    expose_api_raw_anonymous_and_sessionless,
)
from galaxy.webapps.base.controller import (
    BaseAPIController,
    HTTPBadRequest,
)
from tool_shed.dependencies import attribute_handlers
from tool_shed.metadata import repository_metadata_manager
from tool_shed.repository_types import util as rt_util
from tool_shed.util import (
    commit_util,
    encoding_util,
    hg_util,
    metadata_util,
    repository_content_util,
    repository_util,
    tool_util,
)
from tool_shed.webapp import model
from tool_shed.webapp.search.repo_search import RepoSearch

log = logging.getLogger(__name__)


class RepositoriesController(BaseAPIController):
    """RESTful controller for interactions with repositories in the Tool Shed."""

    @web.legacy_expose_api
    def add_repository_registry_entry(self, trans, payload, **kwd):
        """
        POST /api/repositories/add_repository_registry_entry
        Adds appropriate entries to the repository registry for the repository defined by the received name and owner.

        :param key: the user's API key

        The following parameters are included in the payload.
        :param tool_shed_url (required): the base URL of the Tool Shed containing the Repository
        :param name (required): the name of the Repository
        :param owner (required): the owner of the Repository
        """
        response_dict = {}
        if not trans.user_is_admin:
            response_dict["status"] = "error"
            response_dict["message"] = "You are not authorized to add entries to this Tool Shed's repository registry."
            return response_dict
        tool_shed_url = payload.get("tool_shed_url", "")
        if not tool_shed_url:
            raise HTTPBadRequest(detail="Missing required parameter 'tool_shed_url'.")
        tool_shed_url = tool_shed_url.rstrip("/")
        name = payload.get("name", "")
        if not name:
            raise HTTPBadRequest(detail="Missing required parameter 'name'.")
        owner = payload.get("owner", "")
        if not owner:
            raise HTTPBadRequest(detail="Missing required parameter 'owner'.")
        repository = repository_util.get_repository_by_name_and_owner(self.app, name, owner)
        if repository is None:
            error_message = f"Cannot locate repository with name {name} and owner {owner},"
            log.debug(error_message)
            response_dict["status"] = "error"
            response_dict["message"] = error_message
            return response_dict
        # Update the repository registry.
        self.app.repository_registry.add_entry(repository)
        response_dict["status"] = "ok"
        response_dict[
            "message"
        ] = f"Entries for repository {name} owned by {owner} have been added to the Tool Shed repository registry."
        return response_dict

    @web.legacy_expose_api_anonymous
    def get_ordered_installable_revisions(self, trans, name=None, owner=None, **kwd):
        """
        GET /api/repositories/get_ordered_installable_revisions

        :param name: the name of the Repository
        :param owner: the owner of the Repository

        Returns the ordered list of changeset revision hash strings that are associated with installable revisions.
        As in the changelog, the list is ordered oldest to newest.
        """
        # Example URL: http://localhost:9009/api/repositories/get_ordered_installable_revisions?name=add_column&owner=test
        if name is None:
            name = kwd.get("name", None)
        if owner is None:
            owner = kwd.get("owner", None)
        tsr_id = kwd.get("tsr_id", None)
        eagerload_columns = [model.Repository.downloadable_revisions]
        if None not in [name, owner]:
            # Get the repository information.
            repository = repository_util.get_repository_by_name_and_owner(
                self.app, name, owner, eagerload_columns=eagerload_columns
            )
            if repository is None:
                trans.response.status = 404
                return {"status": "error", "message": f"No repository named {name} found with owner {owner}"}
        elif tsr_id is not None:
            repository = repository_util.get_repository_in_tool_shed(
                self.app, tsr_id, eagerload_columns=eagerload_columns
            )
        else:
            error_message = "Error in the Tool Shed repositories API in get_ordered_installable_revisions: "
            error_message += "invalid parameters received."
            log.debug(error_message)
            return []
        return [revision[1] for revision in repository.installable_revisions(self.app, sort_revisions=True)]

    @web.legacy_expose_api_anonymous
    def get_repository_revision_install_info(self, trans, name, owner, changeset_revision, **kwd):
        """
        GET /api/repositories/get_repository_revision_install_info

        :param name: the name of the Repository
        :param owner: the owner of the Repository
        :param changeset_revision: the changeset_revision of the RepositoryMetadata object associated with the Repository

        Returns a list of the following dictionaries

        - a dictionary defining the Repository.  For example::

            {
                "deleted": false,
                "deprecated": false,
                "description": "add_column hello",
                "id": "f9cad7b01a472135",
                "long_description": "add_column hello",
                "name": "add_column",
                "owner": "test",
                "private": false,
                "times_downloaded": 6,
                "url": "/api/repositories/f9cad7b01a472135",
                "user_id": "f9cad7b01a472135"
            }

        - a dictionary defining the Repository revision (RepositoryMetadata).  For example::

            {
                "changeset_revision": "3a08cc21466f",
                "downloadable": true,
                "has_repository_dependencies": false,
                "has_repository_dependencies_only_if_compiling_contained_td": false,
                "id": "f9cad7b01a472135",
                "includes_datatypes": false,
                "includes_tool_dependencies": false,
                "includes_tools": true,
                "includes_tools_for_display_in_tool_panel": true,
                "includes_workflows": false,
                "malicious": false,
                "repository_id": "f9cad7b01a472135",
                "url": "/api/repository_revisions/f9cad7b01a472135",
                "valid_tools": [{u'add_to_tool_panel': True,
                    u'description': u'data on any column using simple expressions',
                    u'guid': u'localhost:9009/repos/enis/sample_repo_1/Filter1/2.2.0',
                    u'id': u'Filter1',
                    u'name': u'Filter',
                    u'requirements': [],
                    u'tests': [{u'inputs': [[u'input', u'1.bed'], [u'cond', u"c1=='chr22'"]],
                    u'name': u'Test-1',
                    u'outputs': [[u'out_file1', u'filter1_test1.bed']],
                    u'required_files': [u'1.bed', u'filter1_test1.bed']}],
                    u'tool_config': u'database/community_files/000/repo_1/filtering.xml',
                    u'tool_type': u'default',
                    u'version': u'2.2.0',
                    u'version_string_cmd': None}]
            }

        - a dictionary including the additional information required to install the repository.  For example::

            {
                "add_column": [
                    "add_column hello",
                    "http://test@localhost:9009/repos/test/add_column",
                    "3a08cc21466f",
                    "1",
                    "test",
                    {},
                    {}
                ]
            }

        """
        # Example URL:
        # http://<xyz>/api/repositories/get_repository_revision_install_info?name=<n>&owner=<o>&changeset_revision=<cr>
        if name and owner and changeset_revision:
            # Get the repository information.
            repository = repository_util.get_repository_by_name_and_owner(
                self.app, name, owner, eagerload_columns=[model.Repository.downloadable_revisions]
            )
            if repository is None:
                log.debug(f"Cannot locate repository {name} owned by {owner}")
                return {}, {}, {}
            encoded_repository_id = trans.security.encode_id(repository.id)
            repository_dict = repository.to_dict(view="element", value_mapper=self.__get_value_mapper(trans))
            repository_dict["url"] = web.url_for(controller="repositories", action="show", id=encoded_repository_id)
            # Get the repository_metadata information.
            repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
                self.app, encoded_repository_id, changeset_revision
            )
            if repository_metadata is None:
                # The changeset_revision column in the repository_metadata table has been updated with a new
                # value value, so find the changeset_revision to which we need to update.
                new_changeset_revision = metadata_util.get_next_downloadable_changeset_revision(
                    self.app, repository, changeset_revision
                )
                repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
                    self.app, encoded_repository_id, new_changeset_revision
                )
                changeset_revision = new_changeset_revision
            if repository_metadata is not None:
                encoded_repository_metadata_id = trans.security.encode_id(repository_metadata.id)
                repository_metadata_dict = repository_metadata.to_dict(
                    view="collection", value_mapper=self.__get_value_mapper(trans)
                )
                repository_metadata_dict["url"] = web.url_for(
                    controller="repository_revisions", action="show", id=encoded_repository_metadata_id
                )
                if "tools" in repository_metadata.metadata:
                    repository_metadata_dict["valid_tools"] = repository_metadata.metadata["tools"]
                # Get the repo_info_dict for installing the repository.
                (
                    repo_info_dict,
                    includes_tools,
                    includes_tool_dependencies,
                    includes_tools_for_display_in_tool_panel,
                    has_repository_dependencies,
                    has_repository_dependencies_only_if_compiling_contained_td,
                ) = repository_util.get_repo_info_dict(self.app, trans.user, encoded_repository_id, changeset_revision)
                return repository_dict, repository_metadata_dict, repo_info_dict
            else:
                log.debug(
                    "Unable to locate repository_metadata record for repository id %s and changeset_revision %s",
                    repository.id,
                    changeset_revision,
                )
                return repository_dict, {}, {}
        else:
            debug_msg = "Error in the Tool Shed repositories API in get_repository_revision_install_info: "
            debug_msg += f"Invalid name {name} or owner {owner} or changeset_revision {changeset_revision} received."
            log.debug(debug_msg)
            return {}, {}, {}

    @web.legacy_expose_api_anonymous
    def get_installable_revisions(self, trans, **kwd):
        """
        GET /api/repositories/get_installable_revisions

        :param tsr_id: the encoded toolshed ID of the repository

        Returns a list of lists of changesets, in the format [ [ 0, fbb391dc803c ], [ 1, 9d9ec4d9c03e ], [ 2, 9b5b20673b89 ], [ 3, e8c99ce51292 ] ].
        """
        # Example URL: http://localhost:9009/api/repositories/get_installable_revisions?tsr_id=9d37e53072ff9fa4
        tsr_id = kwd.get("tsr_id", None)
        if tsr_id is not None:
            repository = repository_util.get_repository_in_tool_shed(
                self.app, tsr_id, eagerload_columns=[model.Repository.downloadable_revisions]
            )
        else:
            error_message = "Error in the Tool Shed repositories API in get_ordered_installable_revisions: "
            error_message += "missing or invalid parameter received."
            log.debug(error_message)
            return []
        return repository.installable_revisions(self.app)

    def __get_value_mapper(self, trans) -> Dict[str, Callable]:
        value_mapper = {
            "id": trans.security.encode_id,
            "repository_id": trans.security.encode_id,
            "user_id": trans.security.encode_id,
        }
        return value_mapper

    @expose_api_raw_anonymous_and_sessionless
    def index(self, trans, deleted=False, owner=None, name=None, **kwd):
        """
        GET /api/repositories
        Displays a collection of repositories with optional criteria.

        :param q:        (optional)if present search on the given query will be performed
        :type  q:        str

        :param page:     (optional)requested page of the search
        :type  page:     int

        :param page_size:     (optional)requested page_size of the search
        :type  page_size:     int

        :param jsonp:    (optional)flag whether to use jsonp format response, defaults to False
        :type  jsonp:    bool

        :param callback: (optional)name of the function to wrap callback in
                         used only when jsonp is true, defaults to 'callback'
        :type  callback: str

        :param deleted:  (optional)displays repositories that are or are not set to deleted.
        :type  deleted:  bool

        :param owner:    (optional)the owner's public username.
        :type  owner:    str

        :param name:     (optional)the repository name.
        :type  name:     str

        :param tool_ids:  (optional) a tool GUID to find the repository for
        :param tool_ids:  str

        :returns dict:   object containing list of results

        Examples:
            GET http://localhost:9009/api/repositories
            GET http://localhost:9009/api/repositories?q=fastq
        """
        repository_dicts = []
        deleted = util.asbool(deleted)
        q = kwd.get("q", "")
        if q:
            page = kwd.get("page", 1)
            page_size = kwd.get("page_size", 10)
            try:
                page = int(page)
                page_size = int(page_size)
            except ValueError:
                raise RequestParameterInvalidException('The "page" and "page_size" parameters have to be integers.')
            return_jsonp = util.asbool(kwd.get("jsonp", False))
            callback = kwd.get("callback", "callback")
            search_results = self._search(trans, q, page, page_size)
            if return_jsonp:
                response = str(f"{callback}({json.dumps(search_results)});")
            else:
                response = json.dumps(search_results)
            return response
        tool_ids = kwd.get("tool_ids", None)
        if tool_ids is not None:
            tool_ids = util.listify(tool_ids)
            repository_found = []
            all_metadata = dict()
            for tool_id in tool_ids:
                # A valid GUID looks like toolshed.g2.bx.psu.edu/repos/bgruening/deeptools/deeptools_computeMatrix/1.1.0
                shed, _, owner, name, tool, version = tool_id.split("/")
                clause_list = [
                    and_(
                        self.app.model.Repository.table.c.deprecated == false(),
                        self.app.model.Repository.table.c.deleted == false(),
                        self.app.model.Repository.table.c.name == name,
                        self.app.model.User.table.c.username == owner,
                        self.app.model.Repository.table.c.user_id == self.app.model.User.table.c.id,
                    )
                ]
                repository = trans.sa_session.query(self.app.model.Repository).filter(*clause_list).first()
                if not repository:
                    log.warning(f"Repository {owner}/{name} does not exist, skipping")
                    continue
                for changeset, changehash in repository.installable_revisions(self.app):
                    metadata = metadata_util.get_current_repository_metadata_for_changeset_revision(
                        self.app, repository, changehash
                    )
                    tools = metadata.metadata.get("tools")
                    if not tools:
                        log.warning(f"Repository {owner}/{name}/{changehash} does not contain valid tools, skipping")
                        continue
                    for tool in tools:
                        if tool["guid"] in tool_ids:
                            repository_found.append("%d:%s" % (int(changeset), changehash))
                    metadata = metadata_util.get_current_repository_metadata_for_changeset_revision(
                        self.app, repository, changehash
                    )
                    if metadata is None:
                        continue
                    metadata_dict = metadata.to_dict(
                        value_mapper={"id": self.app.security.encode_id, "repository_id": self.app.security.encode_id}
                    )
                    metadata_dict["repository"] = repository.to_dict(value_mapper={"id": self.app.security.encode_id})
                    if metadata.has_repository_dependencies:
                        metadata_dict["repository_dependencies"] = metadata_util.get_all_dependencies(
                            self.app, metadata, processed_dependency_links=[]
                        )
                    else:
                        metadata_dict["repository_dependencies"] = []
                    if metadata.includes_tool_dependencies:
                        metadata_dict["tool_dependencies"] = repository.get_tool_dependencies(self.app, changehash)
                    else:
                        metadata_dict["tool_dependencies"] = {}
                    if metadata.includes_tools:
                        metadata_dict["tools"] = metadata.metadata["tools"]
                    all_metadata[f"{int(changeset)}:{changehash}"] = metadata_dict
            if repository_found:
                all_metadata["current_changeset"] = repository_found[0]
                # all_metadata[ 'found_changesets' ] = repository_found
                return json.dumps(all_metadata)
            return "{}"

        clause_list = [
            and_(
                self.app.model.Repository.table.c.deprecated == false(),
                self.app.model.Repository.table.c.deleted == deleted,
            )
        ]
        if owner is not None:
            clause_list.append(
                and_(
                    self.app.model.User.table.c.username == owner,
                    self.app.model.Repository.table.c.user_id == self.app.model.User.table.c.id,
                )
            )
        if name is not None:
            clause_list.append(self.app.model.Repository.table.c.name == name)
        for repository in (
            trans.sa_session.query(self.app.model.Repository)
            .filter(*clause_list)
            .order_by(self.app.model.Repository.table.c.name)
        ):
            repository_dict = repository.to_dict(view="collection", value_mapper=self.__get_value_mapper(trans))
            repository_dict["category_ids"] = [trans.security.encode_id(x.category.id) for x in repository.categories]
            repository_dicts.append(repository_dict)
        return json.dumps(repository_dicts)

    def _search(self, trans, q, page=1, page_size=10):
        """
        Perform the search over TS repositories.
        Note that search works over the Whoosh index which you have
        to pre-create with scripts/tool_shed/build_ts_whoosh_index.sh manually.
        Also TS config option toolshed_search_on has to be True and
        whoosh_index_dir has to be specified.
        """
        conf = self.app.config
        if not conf.toolshed_search_on:
            raise ConfigDoesNotAllowException("Searching the TS through the API is turned off for this instance.")
        if not conf.whoosh_index_dir:
            raise ConfigDoesNotAllowException(
                "There is no directory for the search index specified. Please contact the administrator."
            )
        search_term = q.strip()
        if len(search_term) < 1:
            raise RequestParameterInvalidException("The search term has to be at least one character long.")

        repo_search = RepoSearch()

        Boosts = namedtuple(
            "Boosts",
            [
                "repo_name_boost",
                "repo_description_boost",
                "repo_long_description_boost",
                "repo_homepage_url_boost",
                "repo_remote_repository_url_boost",
                "categories_boost",
                "repo_owner_username_boost",
            ],
        )
        boosts = Boosts(
            float(conf.get("repo_name_boost", 0.9)),
            float(conf.get("repo_description_boost", 0.6)),
            float(conf.get("repo_long_description_boost", 0.5)),
            float(conf.get("repo_homepage_url_boost", 0.3)),
            float(conf.get("repo_remote_repository_url_boost", 0.2)),
            float(conf.get("categories_boost", 0.5)),
            float(conf.get("repo_owner_username_boost", 0.3)),
        )

        results = repo_search.search(trans, search_term, page, page_size, boosts)
        results["hostname"] = web.url_for("/", qualified=True)
        return results

    @web.legacy_expose_api
    def remove_repository_registry_entry(self, trans, payload, **kwd):
        """
        POST /api/repositories/remove_repository_registry_entry
        Removes appropriate entries from the repository registry for the repository defined by the received name and owner.

        :param key: the user's API key

        The following parameters are included in the payload.
        :param tool_shed_url (required): the base URL of the Tool Shed containing the Repository
        :param name (required): the name of the Repository
        :param owner (required): the owner of the Repository
        """
        response_dict = {}
        if not trans.user_is_admin:
            response_dict["status"] = "error"
            response_dict[
                "message"
            ] = "You are not authorized to remove entries from this Tool Shed's repository registry."
            return response_dict
        tool_shed_url = payload.get("tool_shed_url", "")
        if not tool_shed_url:
            raise HTTPBadRequest(detail="Missing required parameter 'tool_shed_url'.")
        tool_shed_url = tool_shed_url.rstrip("/")
        name = payload.get("name", "")
        if not name:
            raise HTTPBadRequest(detail="Missing required parameter 'name'.")
        owner = payload.get("owner", "")
        if not owner:
            raise HTTPBadRequest(detail="Missing required parameter 'owner'.")
        repository = repository_util.get_repository_by_name_and_owner(self.app, name, owner)
        if repository is None:
            error_message = f"Cannot locate repository with name {name} and owner {owner},"
            log.debug(error_message)
            response_dict["status"] = "error"
            response_dict["message"] = error_message
            return response_dict
        # Update the repository registry.
        self.app.repository_registry.remove_entry(repository)
        response_dict["status"] = "ok"
        response_dict[
            "message"
        ] = f"Entries for repository {name} owned by {owner} have been removed from the Tool Shed repository registry."
        return response_dict

    @web.legacy_expose_api
    def reset_metadata_on_repositories(self, trans, payload, **kwd):
        """
        PUT /api/repositories/reset_metadata_on_repositories

        Resets all metadata on all repositories in the Tool Shed in an "orderly fashion".  Since there are currently only two
        repository types (tool_dependecy_definition and unrestricted), the order in which metadata is reset is repositories of
        type tool_dependecy_definition first followed by repositories of type unrestricted, and only one pass is necessary.  If
        a new repository type is introduced, the process will undoubtedly need to be revisited.  To facilitate this order, an
        in-memory list of repository ids that have been processed is maintained.

        :param key: the API key of the Tool Shed user.
        :param my_writable (optional):
            if the API key is associated with an admin user in the Tool Shed, setting this param value
            to True will restrict resetting metadata to only repositories that are writable by the user
            in addition to those repositories of type tool_dependency_definition.  This param is ignored
            if the current user is not an admin user, in which case this same restriction is automatic.

        :param encoded_ids_to_skip (optional): a list of encoded repository ids for repositories that should not be processed.
        :param skip_file (optional):
            A local file name that contains the encoded repository ids associated with repositories to skip.
            This param can be used as an alternative to the above encoded_ids_to_skip.

        """

        def handle_repository(trans, repository, results):
            log.debug(f"Resetting metadata on repository {repository.name}")
            try:
                rmm = repository_metadata_manager.RepositoryMetadataManager(
                    app=self.app,
                    user=trans.user,
                    resetting_all_metadata_on_repository=True,
                    updating_installed_repository=False,
                    repository=repository,
                    persist=False,
                )
                rmm.reset_all_metadata_on_repository_in_tool_shed()
                rmm_invalid_file_tups = rmm.get_invalid_file_tups()
                if rmm_invalid_file_tups:
                    message = tool_util.generate_message_for_invalid_tools(
                        self.app, rmm_invalid_file_tups, repository, None, as_html=False
                    )
                    results["unsuccessful_count"] += 1
                else:
                    message = f"Successfully reset metadata on repository {repository.name} owned by {repository.user.username}"
                    results["successful_count"] += 1
            except Exception as e:
                message = (
                    f"Error resetting metadata on repository {repository.name} owned by {repository.user.username}: {e}"
                )
                results["unsuccessful_count"] += 1
            status = f"{repository.name} : {message}"
            results["repository_status"].append(status)
            return results

        start_time = strftime("%Y-%m-%d %H:%M:%S")
        results = dict(start_time=start_time, repository_status=[], successful_count=0, unsuccessful_count=0)
        handled_repository_ids = []
        encoded_ids_to_skip = payload.get("encoded_ids_to_skip", [])
        skip_file = payload.get("skip_file", None)
        if skip_file and os.path.exists(skip_file) and not encoded_ids_to_skip:
            # Load the list of encoded_ids_to_skip from the skip_file.
            # Contents of file must be 1 encoded repository id per line.
            lines = open(skip_file, "rb").readlines()
            for line in lines:
                if line.startswith("#"):
                    # Skip comments.
                    continue
                encoded_ids_to_skip.append(line.rstrip("\n"))
        if trans.user_is_admin:
            my_writable = util.asbool(payload.get("my_writable", False))
        else:
            my_writable = True
        rmm = repository_metadata_manager.RepositoryMetadataManager(
            app=self.app,
            user=trans.user,
            resetting_all_metadata_on_repository=True,
            updating_installed_repository=False,
            persist=False,
        )
        query = rmm.get_query_for_setting_metadata_on_repositories(my_writable=my_writable, order=False)
        # First reset metadata on all repositories of type repository_dependency_definition.
        for repository in query:
            encoded_id = trans.security.encode_id(repository.id)
            if encoded_id in encoded_ids_to_skip:
                log.debug(
                    "Skipping repository with id %s because it is in encoded_ids_to_skip %s",
                    repository.id,
                    encoded_ids_to_skip,
                )
            elif repository.type == rt_util.TOOL_DEPENDENCY_DEFINITION and repository.id not in handled_repository_ids:
                results = handle_repository(trans, repository, results)
        # Now reset metadata on all remaining repositories.
        for repository in query:
            encoded_id = trans.security.encode_id(repository.id)
            if encoded_id in encoded_ids_to_skip:
                log.debug(
                    "Skipping repository with id %s because it is in encoded_ids_to_skip %s",
                    repository.id,
                    encoded_ids_to_skip,
                )
            elif repository.type != rt_util.TOOL_DEPENDENCY_DEFINITION and repository.id not in handled_repository_ids:
                results = handle_repository(trans, repository, results)
        stop_time = strftime("%Y-%m-%d %H:%M:%S")
        results["stop_time"] = stop_time
        return json.dumps(results, sort_keys=True, indent=4)

    @web.legacy_expose_api
    def reset_metadata_on_repository(self, trans, payload, **kwd):
        """
        POST /api/repositories/reset_metadata_on_repository

        Resets all metadata on a specified repository in the Tool Shed.

        :param key: the API key of the Tool Shed user.

        The following parameters must be included in the payload.
        :param repository_id: the encoded id of the repository on which metadata is to be reset.
        """

        def handle_repository(trans, start_time, repository):
            results = dict(start_time=start_time, repository_status=[])
            try:
                rmm = repository_metadata_manager.RepositoryMetadataManager(
                    app=self.app,
                    user=trans.user,
                    repository=repository,
                    resetting_all_metadata_on_repository=True,
                    updating_installed_repository=False,
                    persist=False,
                )
                rmm.reset_all_metadata_on_repository_in_tool_shed()
                rmm_invalid_file_tups = rmm.get_invalid_file_tups()
                if rmm_invalid_file_tups:
                    message = tool_util.generate_message_for_invalid_tools(
                        self.app, rmm_invalid_file_tups, repository, None, as_html=False
                    )
                    results["status"] = "warning"
                else:
                    message = f"Successfully reset metadata on repository {repository.name} owned by {repository.user.username}"
                    results["status"] = "ok"
            except Exception as e:
                message = (
                    f"Error resetting metadata on repository {repository.name} owned by {repository.user.username}: {e}"
                )
                results["status"] = "error"
            status = f"{repository.name} : {message}"
            results["repository_status"].append(status)
            return results

        repository_id = payload.get("repository_id", None)
        if repository_id is not None:
            repository = repository_util.get_repository_in_tool_shed(self.app, repository_id)
            start_time = strftime("%Y-%m-%d %H:%M:%S")
            log.debug(f"{start_time}...resetting metadata on repository {repository.name}")
            results = handle_repository(trans, start_time, repository)
            stop_time = strftime("%Y-%m-%d %H:%M:%S")
            results["stop_time"] = stop_time
        return results

    @expose_api_anonymous_and_sessionless
    def show(self, trans, id, **kwd):
        """
        GET /api/repositories/{encoded_repository_id}
        Returns information about a repository in the Tool Shed.

        Example URL: http://localhost:9009/api/repositories/f9cad7b01a472135

        :param id: the encoded id of the Repository object
        :type  id: encoded str

        :returns:   detailed repository information
        :rtype:     dict

        :raises:  ObjectNotFound, MalformedId
        """
        repository = repository_util.get_repository_in_tool_shed(self.app, id)
        if repository is None:
            raise ObjectNotFound("Unable to locate repository for the given id.")
        repository_dict = repository.to_dict(view="element", value_mapper=self.__get_value_mapper(trans))
        # TODO the following property would be better suited in the to_dict method
        repository_dict["category_ids"] = [trans.security.encode_id(x.category.id) for x in repository.categories]
        return repository_dict

    @expose_api_raw_anonymous_and_sessionless
    def updates(self, trans, **kwd):
        """
        GET /api/repositories/updates
        Return a dictionary with boolean values for whether there are updates available
        for the repository revision, newer installable revisions available,
        the revision is the latest installable revision, and if the repository is deprecated.

        :param owner: owner of the repository
        :type  owner: str
        :param name: name of the repository
        :type  name: str
        :param changeset_revision: changeset of the repository
        :type  changeset_revision: str
        :param hexlify: flag whether to hexlify the response (for backward compatibility)
        :type  changeset: boolean

        :returns:   information about repository deprecations, updates, and upgrades
        :rtype:     dict
        """
        name = kwd.get("name", None)
        owner = kwd.get("owner", None)
        changeset_revision = kwd.get("changeset_revision", None)
        hexlify_this = util.asbool(kwd.get("hexlify", True))
        repository = repository_util.get_repository_by_name_and_owner(
            trans.app, name, owner, eagerload_columns=[model.Repository.downloadable_revisions]
        )
        if repository and repository.downloadable_revisions:
            repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
                trans.app, trans.security.encode_id(repository.id), changeset_revision
            )
            tool_shed_status_dict = {}
            # Handle repository deprecation.
            tool_shed_status_dict["repository_deprecated"] = str(repository.deprecated)
            tip_revision = repository.downloadable_revisions[0]
            # Handle latest installable revision.
            if changeset_revision == tip_revision:
                tool_shed_status_dict["latest_installable_revision"] = "True"
            else:
                next_installable_revision = metadata_util.get_next_downloadable_changeset_revision(
                    trans.app, repository, changeset_revision
                )
                if repository_metadata is None:
                    if next_installable_revision and next_installable_revision != changeset_revision:
                        tool_shed_status_dict["latest_installable_revision"] = "True"
                    else:
                        tool_shed_status_dict["latest_installable_revision"] = "False"
                else:
                    if next_installable_revision and next_installable_revision != changeset_revision:
                        tool_shed_status_dict["latest_installable_revision"] = "False"
                    else:
                        tool_shed_status_dict["latest_installable_revision"] = "True"
            # Handle revision updates.
            if changeset_revision == tip_revision:
                tool_shed_status_dict["revision_update"] = "False"
            else:
                if repository_metadata is None:
                    tool_shed_status_dict["revision_update"] = "True"
                else:
                    tool_shed_status_dict["revision_update"] = "False"
            # Handle revision upgrades.
            metadata_revisions = [
                revision[1] for revision in metadata_util.get_metadata_revisions(trans.app, repository)
            ]
            num_metadata_revisions = len(metadata_revisions)
            for index, metadata_revision in enumerate(metadata_revisions):
                if index == num_metadata_revisions:
                    tool_shed_status_dict["revision_upgrade"] = "False"
                    break
                if metadata_revision == changeset_revision:
                    if num_metadata_revisions - index > 1:
                        tool_shed_status_dict["revision_upgrade"] = "True"
                    else:
                        tool_shed_status_dict["revision_upgrade"] = "False"
                    break
            return (
                encoding_util.tool_shed_encode(tool_shed_status_dict)
                if hexlify_this
                else json.dumps(tool_shed_status_dict)
            )
        return encoding_util.tool_shed_encode({}) if hexlify_this else json.dumps({})

    @expose_api_anonymous_and_sessionless
    def show_tools(self, trans, id, changeset, **kwd):
        repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(self.app, id, changeset)
        if repository_metadata is not None:
            encoded_repository_metadata_id = trans.security.encode_id(repository_metadata.id)
            repository_metadata_dict = repository_metadata.to_dict(
                view="collection", value_mapper=self.__get_value_mapper(trans)
            )
            repository_metadata_dict["url"] = web.url_for(
                controller="repository_revisions", action="show", id=encoded_repository_metadata_id
            )
            if "tools" in repository_metadata.metadata:
                repository_metadata_dict["valid_tools"] = repository_metadata.metadata["tools"]
            return repository_metadata_dict
        else:
            log.debug(
                "Unable to locate repository_metadata record for repository id %s and changeset_revision %s",
                id,
                changeset,
            )
            return {}

    @expose_api_anonymous_and_sessionless
    def metadata(self, trans, id, **kwd):
        """
        GET /api/repositories/{encoded_repository_id}/metadata
        Returns information about a repository in the Tool Shed.

        Example URL: http://localhost:9009/api/repositories/f9cad7b01a472135/metadata

        :param id: the encoded id of the Repository object

        :param downloadable_only: Return only downloadable revisions (defaults to True).
                                  Added for test cases - shouldn't be considered part of the stable API.

        :returns:   A dictionary containing the specified repository's metadata, by changeset,
                    recursively including dependencies and their metadata.

        :not found:  Empty dictionary.
        """
        recursive = util.asbool(kwd.get("recursive", "True"))
        downloadable_only = util.asbool(kwd.get("downloadable_only", "True"))
        all_metadata = {}
        repository = repository_util.get_repository_in_tool_shed(
            self.app, id, eagerload_columns=[model.Repository.downloadable_revisions]
        )
        for changeset, changehash in metadata_util.get_metadata_revisions(
            self.app, repository, sort_revisions=True, downloadable=downloadable_only
        ):
            metadata = metadata_util.get_current_repository_metadata_for_changeset_revision(
                self.app, repository, changehash
            )
            if metadata is None:
                continue
            metadata_dict = metadata.to_dict(
                value_mapper={"id": self.app.security.encode_id, "repository_id": self.app.security.encode_id}
            )
            metadata_dict["repository"] = repository.to_dict(value_mapper={"id": self.app.security.encode_id})
            if metadata.has_repository_dependencies and recursive:
                metadata_dict["repository_dependencies"] = metadata_util.get_all_dependencies(
                    self.app, metadata, processed_dependency_links=[]
                )
            else:
                metadata_dict["repository_dependencies"] = []
            if metadata.includes_tools:
                metadata_dict["tools"] = metadata.metadata["tools"]
            all_metadata[f"{int(changeset)}:{changehash}"] = metadata_dict
        return all_metadata

    @expose_api
    def update(self, trans, id, **kwd):
        """
        PATCH /api/repositories/{encoded_repository_id}
        Updates information about a repository in the Tool Shed.

        :param id: the encoded id of the Repository object

        :param payload: dictionary structure containing

            'name':                  repo's name (optional)
            'synopsis':              repo's synopsis (optional)
            'description':           repo's description (optional)
            'remote_repository_url': repo's remote repo (optional)
            'homepage_url':          repo's homepage url (optional)
            'category_ids':          list of existing encoded TS category ids the updated repo should be associated with (optional)

        :type payload: dict

        :returns:   detailed repository information
        :rtype:     dict

        :raises: RequestParameterInvalidException, InsufficientPermissionsException
        """
        payload = kwd.get("payload", None)
        if not payload:
            raise RequestParameterMissingException("You did not specify any payload.")

        name = payload.get("name", None)
        synopsis = payload.get("synopsis", None)
        description = payload.get("description", None)
        remote_repository_url = payload.get("remote_repository_url", None)
        homepage_url = payload.get("homepage_url", None)
        category_ids = payload.get("category_ids", None)
        if category_ids is not None:
            # We need to know if it was actually passed, and listify turns None into []
            category_ids = util.listify(category_ids)

        update_kwds = dict(
            name=name,
            description=synopsis,
            long_description=description,
            remote_repository_url=remote_repository_url,
            homepage_url=homepage_url,
            category_ids=category_ids,
        )

        repo, message = repository_util.update_repository(trans, id, **update_kwds)
        if repo is None:
            if "You are not the owner" in message:
                raise InsufficientPermissionsException(message)
            else:
                raise ActionInputError(message)

        repository_dict = repo.to_dict(view="element", value_mapper=self.__get_value_mapper(trans))
        repository_dict["category_ids"] = [trans.security.encode_id(x.category.id) for x in repo.categories]
        return repository_dict

    @expose_api
    def create(self, trans, **kwd):
        """
        POST /api/repositories:

        Creates a new repository.
        Only ``name`` and ``synopsis`` parameters are required.

        :param payload: dictionary structure containing

            'name':                  new repo's name (required)
            'synopsis':              new repo's synopsis (required)
            'description':           new repo's description (optional)
            'remote_repository_url': new repo's remote repo (optional)
            'homepage_url':          new repo's homepage url (optional)
            'category_ids[]':        list of existing encoded TS category ids the new repo should be associated with (optional)
            'type':                  new repo's type, defaults to ``unrestricted`` (optional)

        :type payload: dict

        :returns:   detailed repository information
        :rtype:     dict

        :raises: RequestParameterMissingException, RequestParameterInvalidException
        """
        payload = kwd.get("payload", None)
        if not payload:
            raise RequestParameterMissingException("You did not specify any payload.")
        name = payload.get("name", None)
        if not name:
            raise RequestParameterMissingException("Missing required parameter 'name'.")
        synopsis = payload.get("synopsis", None)
        if not synopsis:
            raise RequestParameterMissingException("Missing required parameter 'synopsis'.")

        description = payload.get("description", "")
        remote_repository_url = payload.get("remote_repository_url", "")
        homepage_url = payload.get("homepage_url", "")
        category_ids = util.listify(payload.get("category_ids[]", ""))

        repo_type = payload.get("type", rt_util.UNRESTRICTED)
        if repo_type not in rt_util.types:
            raise RequestParameterInvalidException("This repository type is not valid")

        invalid_message = repository_util.validate_repository_name(self.app, name, trans.user)
        if invalid_message:
            raise RequestParameterInvalidException(invalid_message)

        repo, message = repository_util.create_repository(
            app=self.app,
            name=name,
            type=repo_type,
            description=synopsis,
            long_description=description,
            user_id=trans.user.id,
            category_ids=category_ids,
            remote_repository_url=remote_repository_url,
            homepage_url=homepage_url,
        )

        repository_dict = repo.to_dict(view="element", value_mapper=self.__get_value_mapper(trans))
        repository_dict["category_ids"] = [trans.security.encode_id(x.category.id) for x in repo.categories]
        return repository_dict

    @web.legacy_expose_api
    def create_changeset_revision(self, trans, id, payload, **kwd):
        """
        POST /api/repositories/{encoded_repository_id}/changeset_revision

        Create a new tool shed repository commit - leaving PUT on parent
        resource open for updating meta-attributes of the repository (and
        Galaxy doesn't allow PUT multipart data anyway
        https://trello.com/c/CQwmCeG6).

        :param id: the encoded id of the Repository object

        The following parameters may be included in the payload.
        :param commit_message: hg commit message for update.
        """

        # Example URL: http://localhost:9009/api/repositories/f9cad7b01a472135
        rdah = attribute_handlers.RepositoryDependencyAttributeHandler(self.app, unpopulate=False)
        tdah = attribute_handlers.ToolDependencyAttributeHandler(self.app, unpopulate=False)

        repository = repository_util.get_repository_in_tool_shed(self.app, id)

        if not (
            trans.user_is_admin
            or self.app.security_agent.user_can_administer_repository(trans.user, repository)
            or self.app.security_agent.can_push(self.app, trans.user, repository)
        ):
            trans.response.status = 400
            return {
                "err_msg": "You do not have permission to update this repository.",
            }

        repo_dir = repository.repo_path(self.app)

        upload_point = commit_util.get_upload_point(repository, **kwd)
        tip = repository.tip()

        file_data = payload.get("file")
        # Code stolen from gx's upload_common.py
        if isinstance(file_data, cgi_FieldStorage):
            assert not isinstance(file_data.file, StringIO)
            assert file_data.file.name != "<fdopen>"
            local_filename = util.mkstemp_ln(file_data.file.name, "upload_file_data_")
            file_data.file.close()
            file_data = dict(filename=file_data.filename, local_filename=local_filename)
        elif type(file_data) == dict and "local_filename" not in file_data:
            raise Exception("Uploaded file was encoded in a way not understood.")

        commit_message = kwd.get("commit_message", "Uploaded")

        uploaded_file = open(file_data["local_filename"], "rb")
        uploaded_file_name = file_data["local_filename"]

        isgzip = False
        isbz2 = False
        isgzip = checkers.is_gzip(uploaded_file_name)
        if not isgzip:
            isbz2 = checkers.is_bz2(uploaded_file_name)
        if isgzip or isbz2:
            # Open for reading with transparent compression.
            tar = tarfile.open(uploaded_file_name, "r:*")
        else:
            tar = tarfile.open(uploaded_file_name)

        new_repo_alert = False
        remove_repo_files_not_in_tar = True

        (
            ok,
            message,
            files_to_remove,
            content_alert_str,
            undesirable_dirs_removed,
            undesirable_files_removed,
        ) = repository_content_util.upload_tar(
            trans,
            rdah,
            tdah,
            repository,
            tar,
            uploaded_file,
            upload_point,
            remove_repo_files_not_in_tar,
            commit_message,
            new_repo_alert,
        )
        if ok:
            # Update the repository files for browsing.
            hg_util.update_repository(repo_dir)
            # Get the new repository tip.
            if tip == repository.tip():
                trans.response.status = 400
                message = "No changes to repository."
                ok = False
            else:
                rmm = repository_metadata_manager.RepositoryMetadataManager(
                    app=self.app, user=trans.user, repository=repository
                )
                status, error_message = rmm.set_repository_metadata_due_to_new_tip(
                    trans.request.host, content_alert_str=content_alert_str, **kwd
                )
                if error_message:
                    ok = False
                    trans.response.status = 500
                    message = error_message
        else:
            trans.response.status = 500
        if os.path.exists(uploaded_file_name):
            os.remove(uploaded_file_name)
        if not ok:
            return {"err_msg": message}
        else:
            return {"message": message}
